/////////////////////////////////////////////////////////////////////////////
// Name:        ArchiveStream.cpp
// Purpose:    Virtual manipulation of archives
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////
// For compilers that support precompilation, includes "wx/wx.h".  
#include "wx/wxprec.h" 

#ifndef WX_PRECOMP
	#include "wx/tokenzr.h"
	#if wxVERSION_NUMBER >= 2504
		#include <wx/stream.h>
		#include <wx/zipstrm.h>
		#include <wx/zstream.h>
		#include <wx/mstream.h>
	#endif
#endif


#include "MyDirs.h"
#include "MyGenericDirCtrl.h"
#include "MyFrame.h"
#include "Externs.h"
#include "Redo.h"
#include "Archive.h"
#include "ArchiveStream.h"
#include "Filetypes.h"

#if wxVERSION_NUMBER >= 2504

	#if wxVERSION_NUMBER < 2702
		#include "tarstream.h"
	#else
		#include "wx/tarstrm.h"
	#endif
	
	#include "bzipstream.h"
#endif
#include <grp.h>

wxString FakeFiledata::PermissionsToText()		// Returns a string describing the filetype & permissions eg -rwxr--r--. Adapted from real FileData, but uses the archivestream::GetMode() info
{
wxString text;
text.Printf(wxT("%c%c%c%c%c%c%c%c%c%c"), IsDir() ? wxT('d') : wxT('-'),	
	!!(Permissions & S_IRUSR) ? wxT('r') : wxT('-'), !!(Permissions & S_IWUSR) ? wxT('w') : wxT('-'), !!(Permissions & S_IXUSR) ? wxT('x') : wxT('-'),
	!!(Permissions & S_IRGRP) ? wxT('r') : wxT('-'), !!(Permissions & S_IWGRP) ? wxT('w') : wxT('-'), !!(Permissions & S_IXGRP) ? wxT('x') : wxT('-'),
	!!(Permissions & S_IROTH) ? wxT('r') : wxT('-'), !!(Permissions & S_IWOTH) ? wxT('w') : wxT('-'), !!(Permissions & S_IXOTH) ? wxT('x') : wxT('-')
				);

if ( text.GetChar(0) == wxT('-') )																// If it wasn't a dir, 
	if ( IsSymlink() ) text.SetChar(0, wxT('l'));											//   see if it's one of the miscellany, rather than a regular file
	else if ( IsCharDev() ) text.SetChar(0, wxT('c'));		
	else if ( IsBlkDev() ) text.SetChar(0, wxT('b'));		
	else if ( IsSocket() ) text.SetChar(0, wxT('s'));		
	else if ( IsFIFO() ) text.SetChar(0, wxT('p'));	

																// Now amend string if the unusual permissions are set.
if ( !!(Permissions & S_ISUID) ) text.GetChar(3) == wxT('x') ?  text.SetChar(3, wxT('s')) : text.SetChar(3, wxT('S'));	//  If originally 'x', use lower case, otherwise upper
if ( !!(Permissions & S_ISGID) ) text.GetChar(6) == wxT('x') ?  text.SetChar(6, wxT('s')) : text.SetChar(6, wxT('S'));
if ( !!(Permissions & S_ISVTX) ) text.GetChar(9) == wxT('x') ?  text.SetChar(9, wxT('t')) : text.SetChar(9, wxT('T'));
			
return text;
}

enum ffscomp FakeFiledata::Compare( FakeFiledata* comparator, FakeFiledata* candidate )	// Finds how the candidate file/dir is related to the comparator one
{
if ( comparator==NULL || candidate==NULL ) return ffsRubbish;
if ( ! comparator->IsDir() )  return ffsRubbish;														// I can't see atm why we'd be comparing a file with another file

wxString candid = candidate->GetFilepath();
if ( ! candidate->IsDir() ) candid = candid.BeforeLast( wxFILE_SEP_PATH ); candid += wxFILE_SEP_PATH;	// In case this is a file, amputate the filename

			// Go thru the dirs segment by segment until there's a mismatch, or 1 runs out.  We can rely on each starting/ending in '/' as we did it in the ctor
wxStringTokenizer candtok( candid.Mid(1), wxFILE_SEP_PATH );
wxStringTokenizer comptok( comparator->GetFilepath().Mid(1), wxFILE_SEP_PATH );
while ( comptok.HasMoreTokens() )
	{	if ( ! candtok.HasMoreTokens() )
			{	if ( candidate->IsDir() ) return ffsParent;											// The candidate is parent to the comparator dir
					else return ffsCousin;																	// Otherwise its a file belonging to a (g)parent dir. ffsCousin is the best description
			}
		wxString tmp = comptok.GetNextToken(); int ans = tmp.Cmp( candtok.GetNextToken() );	// Compare next token of each
		if ( ans != 0 ) return ffsCousin;																	// Different so return that they're not (very) related
	}																														// Otherwise identical so loop
		
if ( candtok.HasMoreTokens() ) return ffsChild;												// The candidate matches but is longer, so it's a (g)child of the comparator
if ( ! candidate->IsDir() ) return ffsChild;															// The candidate matches in terms of Path, but it's a file of the comparator
return ffsEqual;																									// Otherwise they must be identical dirs
}

//-----------------------------------------------------------------------------------------------------------------------

void FakeDir::Clear()
{
for ( int n = (int)dirdataarray->GetCount();  n > 0; --n  )	{	FakeDir* item = dirdataarray->Item(n-1); delete item; }
for ( int n = (int)filedataarray->GetCount();  n > 0; --n  )	{	DataBase* item = filedataarray->Item(n-1); delete item; }
dirdataarray->Clear(); filedataarray->Clear();
}

enum ffscomp FakeDir::AddDir( FakeDir* dir )
{
if ( dir == NULL ) return ffsRubbish;

enum ffscomp ans = Compare( this, dir );															// First compare the dir with this one
if ( ans != ffsChild ) return ans;																			// If it's above us or in another lineage or rubbish, say so

for ( size_t n=0; n < SubDirCount(); ++n )
	{	enum ffscomp ans = Compare( GetFakeDir(n), dir );
		if ( ans == ffsChild )	return GetFakeDir(n)->AddDir( dir );						// It's a (g)child of this dir, add it there
		if ( ans != ffsCousin )	{ delete dir; return ffsDealtwith; } 							// If parent (which shouldn't be possible) or a duplicate or rubbish, abort
	}
				// Still here? Then it must be a previously unacknowledged child. Add it bit by bit, in case it's a grandchild & we haven't yet met its parent
FakeDir *AddHere = this, *lastdir = parentdir;													// The first segment of new dir will be added to us, of course
wxString rest = dir->GetFilepath().Mid( GetFilepath().Len() );							// Get the bit of string subsequent to this dir
wxStringTokenizer extra( rest, wxFILE_SEP_PATH );
while ( extra.HasMoreTokens() )
	{	wxString fp = AddHere->GetFilepath() + extra.GetNextToken() + wxFILE_SEP_PATH;	// Add this segment to the previous path, and create a new child FakeDir with that name
		if ( extra.HasMoreTokens() )																		// If there's more to come after this segment, create a new fakedir for this bit
			{	FakeDir* newdir = new FakeDir( fp, 0, (time_t)0, 0,myNullString,myNullString, lastdir );
				AddHere->AddChildDir( newdir ); parentdir = AddHere; AddHere = newdir;
			}
		 else AddHere->AddChildDir( dir ); 															// Otherwise add the actual new dir!
	}

return ffsDealtwith;
}

enum ffscomp FakeDir::AddFile( FakeFiledata* file )
{
if ( file == NULL ) return ffsRubbish;

enum ffscomp ans = Compare( this, file );															// First compare the file with this one
if ( ans !=  ffsChild ) return ans;																		// If it's above us or in another lineage or rubbish, say so

for ( size_t n=0; n < SubDirCount(); ++n )
	{	enum ffscomp ans = Compare( GetFakeDir(n), file );
		if ( ans == ffsChild )	return GetFakeDir(n)->AddFile( file );					// It belongs in a (g)child of this dir, add it there
		if ( ans != ffsCousin )	{ delete file; return ffsDealtwith; } 						// If it belongs in a parent (which shouldn't be possible) or is a duplicate or rubbish, abort
	}
				// Still here? Then it must be a previously unacknowledged child. Add it bit by bit, in case it's a grandchild & we haven't yet met its parent
wxString rest = file->GetFilepath().Mid( GetFilepath().Len() );							// Get the bit of string subsequent to this dir
wxString filename = rest.AfterLast(wxFILE_SEP_PATH); rest = rest.BeforeLast(wxFILE_SEP_PATH);	// Separate into path & name
	
FakeDir *AddHere = this, *lastdir = parentdir;													// The first segment of any new dir will be added to us, of course
wxStringTokenizer extra( rest, wxFILE_SEP_PATH );										// Go thru any new Path, creating dirs as we go
while ( extra.HasMoreTokens() )
	{	wxString fp = AddHere->GetFilepath() + extra.GetNextToken() + wxFILE_SEP_PATH;	// Add this segment to the previous path, and create a new child FakeDir with that name
		FakeDir* newdir = new FakeDir( fp, 0, (time_t)0, 0,myNullString,myNullString, lastdir );
		AddHere->AddChildDir( newdir ); parentdir = AddHere; AddHere = newdir;
	}

AddHere->AddChildFile( file );			// We've created any necessary dir structure. Now we can finally add the file
return ffsDealtwith;
}

FakeDir* FakeDir::FindSubdirByName( const wxString& dirname )
{
if ( GetFilepath() == dirname || GetFilepath() == dirname+wxFILE_SEP_PATH ) return this;								// If we're it, return us
for ( size_t n=0; n < SubDirCount(); ++n )
	{ FakeDir* child = dirdataarray->Item(n)->FindSubdirByName(dirname);  if ( child != NULL ) return child; }	// Otherwise recurse thru children

return NULL; 
}

FakeFiledata* FakeDir::FindFileByName( wxString filepath )
{
if ( filepath.IsEmpty() )  return NULL;
if ( filepath.Last() == wxFILE_SEP_PATH ) filepath = filepath.BeforeLast(wxFILE_SEP_PATH);	// In case there's a spurious terminal '/'

wxString path = filepath.BeforeLast(wxFILE_SEP_PATH), filename = filepath.AfterLast(wxFILE_SEP_PATH);
FakeDir* dir = FindSubdirByName( path );
if ( dir == NULL ) return NULL;

return dir->GetFakeFile( filename );
}

bool FakeDir::HasSubDirs( wxString& dirname )		// Find a FakeDir with this name, and see if it has subdirs
{
FakeDir* dir = FindSubdirByName( dirname );													// Find if a dir with this name exists. NULL if not
if ( dir != NULL ) return (dir->SubDirCount() > 0);												// Then just count the subdirs
return false;
}

bool FakeDir::HasFiles( wxString& dirname )			// Find a FakeDir with this name, and see if it has files
{
FakeDir* dir = FindSubdirByName( dirname );
if ( dir != NULL ) return (dir->FileCount() > 0);
return false;
}

bool FakeDir::GetFilepaths( wxArrayString& array )
{
if ( ! FileCount() ) return false;

for ( size_t n=0; n < FileCount(); ++n )
	{	DataBase* temp = GetFakeFiledata(n);
		array.Add( temp->GetFilepath() );
	}
return true;
}

void FakeDir::GetDescendants( wxArrayString& array )	// Fill the array with all its files & (recursively) its subdirs
{
GetFilepaths( array );														// First add this dir's files to the array

for ( size_t n=0; n < SubDirCount(); ++n )						// Now for every subdir
	{	FakeDir* temp = GetFakeDir(n); if ( temp==NULL ) continue;
		array.Add( temp->GetFilepath() );							// Add the subdir
		
		temp->GetDescendants( array );							// and recurse thru it
	}
}

unsigned long FakeDir::GetSize()		// Returns the size of its files
{
unsigned long totalsize = 0;

for ( size_t n=0; n < FileCount(); ++n )							// First add the size of each file
	totalsize += GetFakeFiledata(n)->Size();

return totalsize;
}

unsigned long FakeDir::GetTotalSize()		// Returns the size of all its files & those of its subdirs
{
unsigned long totalsize = GetSize();							// First the files
for ( size_t n=0; n < SubDirCount(); ++n )						// Now recursively for every subdir
	totalsize += GetFakeDir(n)->GetTotalSize();

return totalsize;
}
//-----------------------------------------------------------------------------------------------------------------------

FakeFilesystem::FakeFilesystem( wxString filepath, size_t size, time_t Time, size_t Perms, wxString Owner/*=wxT("")*/, wxString Group/*=wxT("")*/ )
{
rootdir = new FakeDir( filepath, size, Time, Perms, Owner, Group );					// Create the 'root' dir, which will be the archive itself
currentdir = rootdir; 																							// We start with the root
}

void FakeFilesystem::AddItem( wxString filepath, size_t size, time_t Time, size_t Perms, DB_filetype Type, wxString& Target/*=wxT("")*/, wxString Owner/*=wxT("")*/, wxString Group/*=wxT("")*/ )
{
if ( filepath.IsEmpty() ) return;

if ( Type == DIRTYPE )		rootdir->AddDir( new FakeDir( filepath, size, Time, Perms, Owner, Group, rootdir ) );
  else 			rootdir->AddFile( new FakeFiledata( filepath, size, Time, Type, Target, Perms, Owner, Group ) );
}

//-----------------------------------------------------------------------------------------------------------------------

bool ArcDir::Open( wxString dirName )
{
#if wxVERSION_NUMBER >= 2504
if ( arc == NULL || arc->Getffs() == NULL || dirName.IsEmpty() ) return false;
if ( arc->Getffs()->GetRootDir()->FindSubdirByName( dirName ) == NULL )  return false;
filepaths.Clear();
startdir = dirName;											// This is the dir to investigate. We need to save it to use in ArcDir::GetFirst
isopen = true; return true;
#else
return false;
#endif
}

bool ArcDir::GetFirst(wxString* filename, const wxString& filespec, int flags)
{
#if wxVERSION_NUMBER >= 2504
SetFileSpec(filespec); SetFlags(flags);

wxString currentcurrentdir = arc->Getffs()->GetCurrentDir()->GetFilepath();			// Save the current currentdir
if ( ! arc->SetPath( startdir ) ) { arc->SetPath(currentcurrentdir); return false; }	// Change it to point to the dir we're interested in

if ( m_flags & wxDIR_DIRS )
	{	wxArrayString temp; arc->GetDirs( temp );
		filepaths = temp;
	}

if ( m_flags & wxDIR_FILES )	arc->Getffs()->GetCurrentDir()->GetFilepaths( filepaths );

arc->SetPath(currentcurrentdir); 																			// Revert to the original currentdir

if ( filepaths.GetCount() > 0 ) { nextfile=0; return GetNext(filename); }
 else return false;
#else
return false;
#endif
}

bool ArcDir::GetNext(wxString* filename)
{
#if wxVERSION_NUMBER >= 2504
if ( filepaths.IsEmpty() || nextfile < 0 ) return false;

while ( (size_t)nextfile < filepaths.GetCount() )
	{	wxString name = filepaths[ nextfile++ ];
		if ( m_filespec.IsEmpty()  																				// An empty filespec is considered to match everything
					||  wxMatchWild( m_filespec, name ) )													// Pass this name thru the filter
			{	*filename = name; return true; }
	}

return false;																											// No (more) names that match filespec
#else
return false;
#endif
}

//-----------------------------------------------------------------------------------------------------------------------
  // We *could* use zipstreams from 2.5.3 (and plain tars anytime), but bzip streams need wxOutputStream::Close which starts in 2.5.4
#if wxVERSION_NUMBER >= 2504

	#include "wx/wfstream.h"

ArchiveStream::ArchiveStream( DataBase* archive )	// It's a real archive
{
m_membuf = NULL;
wxString name = archive->GetFilepath(); if ( name.IsEmpty() ) return;
if ( name.GetChar( name.Len()-1) != wxFILE_SEP_PATH ) name << wxFILE_SEP_PATH;	// Although it's an archive, not a real dir, for our purposes we need it to pretend
ffs = new FakeFilesystem( name, archive->Size(), archive->ModificationTime(), archive->GetPermissions(), archive->GetOwner(), archive->GetGroup() );
Dirty = Nested = false; archivename = name;
}

ArchiveStream::ArchiveStream( wxString name )	// Used for nested archives
{
m_membuf = NULL;
if ( name.IsEmpty() ) return;
if ( name.GetChar( name.Len()-1) != wxFILE_SEP_PATH ) name << wxFILE_SEP_PATH;	// Although it's an archive, not a real dir, for our purposes we need it to pretend
ffs = new FakeFilesystem( name, 0, 0, 0 );
Dirty = false; Nested = true; archivename = name;
}

bool ArchiveStream::LoadToBuffer( wxString filepath )	// Used by the subclassed ctors to load the archive into m_membuf
{
wxFFileInputStream in( filepath );									// Load the archive thru a wxFFileInputStream
wxStreamBuffer streambuf( in, wxStreamBuffer::read );	//  and store it first in a wxStreamBuffer (since this can cope with a stream feed)
size_t size = in.GetSize();												// This is how much was loaded i.e. the file size
if ( ! size ) return false;

wxChar* buf = new wxChar[ size ];								// It's ugly, but the easiest way I've found to get the data from the streambuffer to a memorybuffer
streambuf.Read( buf, size );											//   is by using an intermediate char[]

delete m_membuf;															// Get rid of any old data, otherwise we'll leak when reloading
m_membuf = new wxMemoryBuffer( size );					// Now (at last) we can store the data in a memory buffer
m_membuf->AppendData( buf, size );
delete[] buf;

return true;
}

bool ArchiveStream::LoadToBuffer( wxMemoryBuffer* mbuf )	// Reload m_membuf with data from a different pane's buffer
{
if ( mbuf == NULL ) return false;

delete m_membuf;															// Get rid of any old data, otherwise we'll leak
m_membuf = new wxMemoryBuffer( mbuf->GetDataLen() );
m_membuf->AppendData( mbuf->GetData(), mbuf->GetDataLen() );
return true;
}

bool ArchiveStream::SaveBuffer()		// Saves the compressed archive in buffer back to the filesystem
{
if ( IsNested() ) { Dirty = true; return true; }					// Except we don't if this is a nested archive: it gets saved on Pop

wxFFileOutputStream fileoutstream( archivename ); if ( ! fileoutstream.Ok() ) return false;
fileoutstream.Write( m_membuf->GetData(), m_membuf->GetDataLen() );

return fileoutstream.IsOk();											// Returns true if the Write was successful
}

void ArchiveStream::ListContentsFromBuffer( wxString archivename )
{
wxBusyCursor busy;
GetFromBuffer();
ListContents( archivename );
}

bool ArchiveStream::IsWithin( wxString filepath )	// Find if the passed filepath is within this archive
{
bool isdir = Getffs()->GetRootDir()->FindSubdirByName( filepath ) != NULL;	// Start by seeing if there's a dir with this name
if ( isdir ) return true;
return (Getffs()->GetRootDir()->FindFileByName( filepath ) != NULL);			// Failing that, try for a file
}


FakeFiledataArray* ArchiveStream::GetFiles()	 // Returns the files for the current 'path'
{
return ffs->GetFiles();
}

bool ArchiveStream::GetDirs( wxArrayString& dirs )	 // Returns in the arraystring the files for the current 'path'
{
dirs.Clear();

for ( size_t n=0; n < ffs->GetCurrentDir()->SubDirCount(); ++n )
	{	FakeDir* dir =  ffs->GetCurrentDir()->GetFakeDir( n  );
		dirs.Add( dir->GetFilepath() );				// NB a FakeDir GetFilepath() returns the full filepath, whereas a FakeFiledata GetFilepath() currently just gives the filename
	}
return dirs.GetCount() > 0;
}

bool ArchiveStream::GetDescendants( wxString dirName, wxArrayString& filepaths )	 // Returns in the array all files, dirs & subdirs files for the path
{
filepaths.Clear();

FakeDir* dir = Getffs()->GetRootDir()->FindSubdirByName( dirName );
if (  dir == NULL ) return false;									// Path must be a dir, not a file or a null

dir->GetDescendants( filepaths );								// This does the real work

return filepaths.GetCount() > 0;
}

bool ArchiveStream::SetPath( const wxString& dir )
{
if ( dir.IsEmpty() ) return false;

if ( (dir+wxFILE_SEP_PATH)  == ffs->GetRootDir()->GetFilepath() )  { ffs->SetCurrentDir( ffs->GetRootDir() ); return true; }		// First check that we're not reverting to root

wxString path( dir );
if ( path.GetChar( path.Len()-1) != wxFILE_SEP_PATH ) path << wxFILE_SEP_PATH;

FakeDir* newdir =  ffs->GetRootDir()->FindSubdirByName( path );
if ( newdir != NULL ) { ffs->SetCurrentDir( newdir ); return true; }						// Bingo! There's a subdir with this name

wxString prepath = dir.BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;	// It wasn't a subdir. Let's check & see if it's a file
newdir =  ffs->GetRootDir()->FindSubdirByName( prepath );
if ( newdir == NULL ) return false;																		// No valid parent dir either
if ( newdir->HasFileCalled( dir.AfterLast(wxFILE_SEP_PATH) ) )							// See if this parent dir contains the correct file
	{	ffs->SetCurrentDir( newdir ); return true;	}													// If so, set path to the parent dir

return false;
}


bool ArchiveStream::OnExtract()	 // On Extract from context menu. Finds what to extract
{
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane(); if ( active==NULL ) return false;
wxArrayString filepaths; size_t count = active->GetMultiplePaths( filepaths ); if ( ! count ) return false;

UnRedoManager::StartCluster();
wxArrayString resultingfilepaths; bool result = DoExtract( filepaths, wxEmptyString, resultingfilepaths );
UnRedoManager::EndCluster();
return result;
}

bool ArchiveStream::DoExtract( wxArrayString& filepaths, wxString destination, wxArrayString& resultingfilepaths, bool DontUndo/*=false*/ )	 // From OnExtract or D'n'D/Paste, or Move
{
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane(); if ( active==NULL ) return false;

size_t count = filepaths.GetCount(); if ( ! count ) return false;
wxArrayString WithinArcNames;
for ( size_t n=0; n < count; ++n )
	{	wxString filepath = filepaths[n];																					// For each filepath, find and store the bit distal to the archive name
		if ( ! SortOutNames( filepath ) ) return false;
		
		wxString WithinArcName( WithinArchiveName ); WithinArcNames.Add( WithinArcName );
	}

wxString suggestedfpath = active->arcman->GetPathOutsideArchive();

while ( destination.IsEmpty() )			// Get the desired destination from the user, if it's not already been provided
	{	wxString msg; if ( count >1 ) msg = _("To which directory would you like these files extracted?");
		  else msg = _("To which directory would you like this extracted?");
		destination = wxGetTextFromUser( msg, _("Extracting from archive"), suggestedfpath );
		if ( destination.IsEmpty() ) return false;							// A blank answer must be a hint ;)
	
		wxString destinationpath( destination ); if ( destinationpath.Last() == wxFILE_SEP_PATH ) destinationpath = destinationpath.BeforeLast(wxFILE_SEP_PATH);
		FileData DestinationFD( destinationpath );
		if ( DestinationFD.IsValid() && ! DestinationFD.CanTHISUserWriteExec() )				// Make sure we have permission to write to the dir
			{	wxMessageDialog dialog(MyFrame::mainframe->GetActivePane(),
										_("I'm afraid you don't have permission to Create in this directory\n                       Try again?"), _("No Entry!"), wxYES_NO | wxICON_ERROR );
				if ( dialog.ShowModal() == wxID_YES ) { destination.Empty(); continue; }
				  else return false;
			}
		bool flag=false;
		for ( size_t n=0; n < count; ++n )
			{	FileData fd( destinationpath + wxFILE_SEP_PATH + WithinArcNames[n] );		// Check for pre-existence of each file: at least those in the base dir
				if ( fd.IsValid() )
					{	wxString msg; msg.Printf(_("Sorry, %s already exists\n            Try again?"), fd.GetFilepath().c_str() );
						wxMessageDialog dialog( active, msg, _("Oops!"), wxYES_NO | wxICON_ERROR );
						if ( dialog.ShowModal() == wxID_YES ) { destination.Empty(); flag=true; break; }
						  else return false;
					}
			}
		if ( flag = true) continue;
	}

if ( destination.Last() != wxFILE_SEP_PATH )	destination << wxFILE_SEP_PATH;		// Nasty things would happen otherwise

if ( ! Extract( filepaths, destination, resultingfilepaths ) ) 	return false;

wxArrayInt IDs; IDs.Add( active->GetId() );
if ( ! DontUndo )
	{	UnRedoExtract* UnRedoptr = new UnRedoExtract( resultingfilepaths, destination, IDs );
		UnRedoManager::AddEntry( UnRedoptr );	
	}
	
MyFrame::mainframe->OnUpdateTrees( destination, IDs );
return true;
}

bool ArchiveStream::SortOutNames( wxString& filepath )		// Subroutine used in Extract etc, so made into a function in the name of code reuse
{
WithinArchiveName.Empty(); WithinArchiveNames.Clear(); WithinArchiveNewNames.Clear(); IsDir = false;
archivename = ffs->GetRootDir()->GetFilepath(); if ( archivename.IsEmpty() ) return false;	// Now get the true archivename: the archive expects it ;)

wxString FakeFilesystemPath( ffs->GetCurrentDir()->GetFilepath() ); if ( FakeFilesystemPath.IsEmpty() ) return false;	// First use current path, in case we're recursing in a subdir
wxString filename; 
if ( ! (filepath == archivename.BeforeLast(wxFILE_SEP_PATH) 
					|| filepath.StartsWith( FakeFilesystemPath, &filename )) )								// This effectively removes the current within-archive path from filepath, leaving the filename
	if ( ! filename.IsEmpty() )																									// just in case filepath IS the archive
		{	if ( ffs->GetCurrentDir()->HasDirCalled( filename ) ) IsDir = true;							// We'll need to know, as dirs get a terminal '/'
				else if (  ! ffs->GetCurrentDir()->HasFileCalled( filename ) ) 	return false;		// If it's not a file either, abort!
		}

 															// WithinArchiveName is the filename to extract: may be subpath/filename within the archive
if ( ! filepath.StartsWith( archivename, &WithinArchiveName ) )	return false;					// Get the "filename" bit: may have a bit of path-within-archive too
if ( IsDir )  WithinArchiveName << wxFILE_SEP_PATH;														// Dirs in the archive will have a terminal '/'

archivename = archivename.BeforeLast(wxFILE_SEP_PATH);											// Remove the archive's spurious terminal '/'

return true;
}

bool ArchiveStream::AdjustFilepaths(  wxArrayString& filepaths, wxString& destpath, wxArrayString& newnames, adjFPs whichtype/*=afp_neither*/ )		// Get the within-archive names, any descendants, and (if renaming) the new names
{
WithinArchiveNames.Clear(); WithinArchiveNewNames.Clear();
archivename = ffs->GetRootDir()->GetFilepath(); if ( archivename.IsEmpty() ) return false;	// Get the archive name, which we need to remove
											// If we're from Rename or Extract, newnames will hold the new names/destination path
if ( whichtype==afp_rename && newnames.GetCount() > filepaths.GetCount() ) return false;	// If there are too many of them, abort

for ( size_t n=0; n < filepaths.GetCount(); ++n )																// For every file/dir we were passed
	{	if ( ! filepaths[n].StartsWith( archivename, &WithinArchiveName ) )	continue;		// Amputate the /path/to/archive bit
		wxString WithinArchiveNewname;
		if ( whichtype==afp_rename ) 
			if ( ! newnames[n].StartsWith( archivename, &WithinArchiveNewname ) )	continue;	// We need to get rid of the path from newname too

		FakeDir* fd = ffs->GetRootDir()->FindSubdirByName( filepaths[n] );							// See if it's a dir
		if ( fd == NULL )
			{	FakeFiledata* ffd = ffs->GetRootDir()->FindFileByName( filepaths[n] ); if ( ffd == NULL ) continue;	// If it's not a dir, or a file, abort this one
				WithinArchiveNames.Add( WithinArchiveName );
				if ( whichtype==afp_rename ) 	WithinArchiveNewNames.Add( WithinArchiveNewname );		// If we're renaming, add the corresponding new name
					else { newnames.Add( destpath + filepaths[n].AfterLast(wxFILE_SEP_PATH) ); WithinArchiveNewNames.Add( newnames[n] ); } // Extract to here
			}
		 else								// It's a dir. We need to find & process all descendants too
		 	{	WithinArchiveNames.Add( WithinArchiveName + wxFILE_SEP_PATH );			// but first do the dir itself
				if ( whichtype==afp_rename ) 	WithinArchiveNewNames.Add( WithinArchiveNewname );		// If we're renaming, add the corresponding new dir name
				  else 
					{	wxString dirname;
						if ( filepaths[n].Last() == wxFILE_SEP_PATH ) dirname = (filepaths[n].BeforeLast(wxFILE_SEP_PATH)).AfterLast(wxFILE_SEP_PATH); 
							else dirname = filepaths[n].AfterLast(wxFILE_SEP_PATH);
						newnames.Add( destpath + dirname ); WithinArchiveNewNames.Add( destpath + dirname );   // Extract to here
					}
					
				wxArrayString fpaths; fd->GetDescendants( fpaths );											// This recursively gets all files & subdirs into fpaths
				size_t distalbit = 0;																								// We need this to extract bar/ from archive/foo/bar. Assume 0 to start with, which would be true for foo
				if ( WithinArchiveName.find( wxFILE_SEP_PATH ) != wxString::npos )
					distalbit = WithinArchiveName.BeforeLast(wxFILE_SEP_PATH).Len() + (archivename.Last()==wxFILE_SEP_PATH);	// This gets bar/ from archive/foo/bar
				for ( size_t i=0; i < fpaths.GetCount(); ++i )
					{	wxString descendantpart;
						if ( ! fpaths[i].StartsWith( archivename, &descendantpart ) )	continue;	// For every descendant, get & add the descendant part
						WithinArchiveNames.Add( descendantpart );
						if ( whichtype==afp_rename ) 																	// If we're renaming, construct the corresponding new name for each
								WithinArchiveNewNames.Add( WithinArchiveNewname + descendantpart.Mid( WithinArchiveName.Len() ) );
						 else { newnames.Add( destpath + descendantpart.Mid( distalbit ) ); WithinArchiveNewNames.Add( destpath + descendantpart.Mid( distalbit ) );  } // distalbit is calculated to use just the desired distal segments 
					}
			}
	}

archivename = archivename.BeforeLast(wxFILE_SEP_PATH);											// Remove the archive's spurious terminal '/'
return WithinArchiveNames.GetCount() > 0;
}

bool ArchiveStream::Alter( wxArrayString& filepaths, enum alterarc DoWhich, wxArrayString& newnames )		// This might be add remove del or rename, depending on the enum
{
if ( filepaths.IsEmpty() ) return false;

if ( DoWhich != arc_add )													// If not arc_add, remove the /pathtoarchive/archive bit of each filepath
	{	wxString Unused; if ( ! AdjustFilepaths( filepaths, Unused, newnames,  DoWhich==arc_rename || DoWhich==arc_dup ? afp_rename : afp_neither  ) ) return false; }
 else																					// If arc_add, get the target dir for insertion into WithinArchivename.  I've overloaded newnames to contain its full filepath
 	{	if ( newnames.IsEmpty() )   return false; 
		wxString dest( newnames[0] );
		if ( SortOutNames( dest ) )											 // Sortoutnames also fills archivename. Don't test for success: if we're pasting directly onto the archive, it'll return false
			if ( ! IsDir ) WithinArchiveName = WithinArchiveName.BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;	// If we're pasting onto a file, remove the filename
	}

wxBusyCursor busy;
if ( ! DoAlteration(  filepaths, DoWhich ) ) return false;		// Do the exciting bit in a tar or zip-specific method

size_t size = OutStreamPtr->GetSize();								// This is how much was loaded i.e. the file size
wxChar* buf = new wxChar[ size ];									// It's ugly, but the easiest way I've found to get the data from the stream to a memorybuffer
((wxMemoryOutputStream*)OutStreamPtr.get())->CopyTo( buf, size );	// is by using an intermediate char[]
OutStreamPtr.reset();															// Delete memoutstream

delete m_membuf; m_membuf = new wxMemoryBuffer( size ); // I've no idea why, but trying to delete the old data by SetDataLen( 0 ) failed :(
m_membuf->AppendData( buf, size );									// Store the data
delete[] buf;

SaveBuffer();

ffs->Clear();																			// Throw out the old dir structure, and reload
ListContentsFromBuffer( archivename );

return true;
}

bool ArchiveStream::MovePaste( MyGenericDirCtrl* originctrl, MyGenericDirCtrl* destctrl, wxArrayString& filepaths, wxString& destpath, bool Moving/*=true*/ )	
{
if ( originctrl == NULL || destctrl == NULL || filepaths.IsEmpty() || destpath.IsEmpty() ) return false;

bool PastingFromTrash = filepaths[0].StartsWith( DirectoryForDeletions::GetDeletedName() );	// We'll need this if we'd Cut from an archive ->trash; & now we're Pasting it
bool FromArchive = ! PastingFromTrash && (originctrl->arcman != NULL && originctrl->arcman->IsArchive()); // If we're PastingFromTrash, originctrl->->IsArchive() would be irrelevantly true 
bool ToArchive = (destctrl->arcman != NULL && destctrl->arcman->IsArchive());

wxString OriginPath( filepaths[0] );																		// We'll need this later
OriginPath = OriginPath.BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;
	
wxArrayInt IDs; IDs.Add( destctrl->GetId() );														// Make an int array, & store the ID of destination pane
if ( Moving ) IDs.Add( originctrl->GetId() );															// If Moving, add the origin ID

if ( FromArchive && ! ToArchive )
	{	FileData dp( destpath ); if ( ! dp.IsValid() ) return false;							// Avoid pasting onto a file
		if ( ! dp.IsDir() && ! (dp.IsSymlink() && dp.IsSymlinktargetADir()) )
			destpath = dp.GetPath();

		FileData DestinationFD( destpath );
		if ( ! DestinationFD.CanTHISUserWriteExec() )											// Make sure we have permission to write to the destination dir
			{	wxMessageDialog dialog(destctrl, _("I'm afraid you don't have Permission to write to this Directory"), _("No Entry!"), wxOK | wxICON_ERROR );
				dialog.ShowModal(); return false;
			}

		enum DupRen WhattodoifClash = Unknown, WhattodoifCantRead = Unknown;		// These get passed to CheckDupRen.  May become SkipAll, OverwriteAll etc, but start with Unknown
		size_t filecount = filepaths.GetCount(); wxArrayString fpaths;	// fpaths will hold the filepaths once they've been checked/amended etc
		for ( size_t n=0; n < filecount; ++n )											// For every entry in the clipboard
			{	bool ItsADir = false;															// We need to check we're not about to do something illegal
				FakeFiledata* fd = Getffs()->GetRootDir()->FindFileByName( filepaths[n] );					// Dirs aren't '/'-terminated, so just try it and see
				if ( fd==NULL )  
					{ fd = Getffs()->GetRootDir()->FindSubdirByName( filepaths[n] ); ItsADir = true; }	// Not a file, so try for a dir
				if ( fd==NULL ) continue;
				CheckDupRen CheckIt(destctrl, filepaths[n], destpath, true, fd->ModificationTime());	// Make an instance of the class that tests for this. The 'true' means "From archive"
				CheckIt.IsMultiple = ( (n+1) < filecount );							// This flags whether there'll be more items after this one, so use different dialogs
				CheckIt.WhattodoifClash = WhattodoifClash; CheckIt.WhattodoifCantRead = WhattodoifCantRead; CheckIt.ItsADir = ItsADir;	// Set CheckIt's variables to match any previous choice
				if ( ! CheckIt.CheckForPreExistence() ) 								// Check we're not pasting onto ourselves or a namesake. True means OK
					{	WhattodoifClash = CheckIt.WhattodoifClash;				// Store any new choice in local variable.  This catches SkipAll & Cancel, which return false
						if ( WhattodoifClash==Cancel ) 	{ n = filecount; break; }	// Cancel this item & the rest of the loop by increasing n
						 else continue; 															// Skip this item but continue with loop
					}
				 else 
						fpaths.Add( filepaths[n] );	// If there was an unskipped clash, this will effect an overwrite. Rename isn't possible ex archive (ie it isn't worth the bother of altering everything to implement it)
			}
		filepaths = fpaths;																		// Revert to using filepaths. This is a reference, so the caller can use it to count the items actually done
		if ( ! filepaths.GetCount() ) return false;									// If everything was skipped/cancelled, go home

		bool NeedClustering = ! UnRedoManager::ClusterIsOpen;
		if ( NeedClustering ) UnRedoManager::StartCluster();
		wxArrayString destinations;
		originctrl->arcman->GetArc()->DoExtract( filepaths, destpath, destinations );
		
		if ( Moving )																					// If we're Moving (cf Pasting), delete those files from the archive
			{	wxArrayString unused; originctrl->arcman->GetArc()->Alter( filepaths, arc_remove, unused );
				UnRedoMoveToFromArchive* UnRedoptr = new UnRedoMoveToFromArchive( filepaths, destinations, OriginPath, destpath, IDs, originctrl, destctrl, amt_from  );
				UnRedoManager::AddEntry( UnRedoptr );
			}

		if ( NeedClustering ) UnRedoManager::EndCluster(); 
		return true;
	}

if ( ToArchive && ! FromArchive )
	{	wxArrayString fpaths, destinations;
		for ( size_t n=0; n < filepaths.GetCount(); ++n )										// We need to go thru the filepaths to add the children of any dirs
			{	fpaths.Add( filepaths.Item(n) );
				RecursivelyGetFilepaths( filepaths.Item(n), fpaths );
			}

		FakeFiledata* ffd = destctrl->arcman->GetArc()->ffs->GetRootDir()->FindFileByName( destpath );	// Destpath doesn't have a '/', so we need to see if it's a file and amputate
		if ( ffd != NULL ) destpath = destpath.BeforeLast(wxFILE_SEP_PATH);
		if ( destpath.Last() != wxFILE_SEP_PATH ) destpath << wxFILE_SEP_PATH;

		wxArrayString Path; Path.Add( destpath ); 												// We have to provide a 2nd arraystring anyway due to the Alter API, so use it to transport the path
		if ( ! destctrl->arcman->GetArc()->Alter( fpaths, arc_add, Path ) ) return false;	// Do the Paste bit of the Move
						
		for ( size_t n=0; n < filepaths.GetCount(); ++n )										// Do the delete bit
			{	if ( Moving )
					{	wxFileName From( filepaths[n] );
						originctrl->ReallyDelete( &From );
					}	
				wxString distal; if ( ! filepaths[n].StartsWith( OriginPath, &distal ) )	continue;
				destinations.Add( destpath + distal );						// Fill destinations array ready for UnRedoing
			}
		
		if ( Moving )
			{	UnRedoMoveToFromArchive* UnRedoptr = new UnRedoMoveToFromArchive( filepaths, destinations, OriginPath, destpath, IDs, originctrl, destctrl, amt_to  );
				UnRedoManager::AddEntry( UnRedoptr );	
				MyFrame::mainframe->OnUpdateTrees( OriginPath, IDs );
			}
		 else											// Move would have done the Paste Unredo bit too. Since we're not Moving, do it here
			{	UnRedoArchivePaste* UnRedoptr = new UnRedoArchivePaste( fpaths, destinations, destpath, IDs, destctrl );
				UnRedoManager::AddEntry( UnRedoptr );
			}
	
		MyFrame::mainframe->OnUpdateTrees( destpath, IDs );
		return true;
	}

	// If we're here, we are moving/pasting from one archive to another, which might be:
	//		Within the same archive in the same pane						--- Rename/Dup
	//		Within the same archive displayed in different panes		--- Rename/Dup, then reload the other one (as otherwise the 2 will be out of sync)
	//		Between different archives displayed in different panes	--- Extract to a tempfile, then add to the 2nd archive. If moving, delete from the 1st
if ( originctrl->arcman->GetArc()->Getffs()->GetRootDir()->GetFilepath() 			//	Within the same archive in the same or different panes, so Rename
																			== destctrl->arcman->GetArc()->Getffs()->GetRootDir()->GetFilepath() )
	{	wxArrayString newpaths;		
		FakeFiledata* ffd = ffs->GetRootDir()->FindFileByName( destpath );		// Destpath doesn't have a '/', so we need to see if it's a file and amputate. The '/' gets added later
		if ( ffd != NULL ) destpath = destpath.BeforeLast(wxFILE_SEP_PATH);
		
		if ( destpath.Last() != wxFILE_SEP_PATH ) destpath << wxFILE_SEP_PATH;
		for ( size_t n=0; n < filepaths.GetCount(); ++n )										// Go thru the filepaths, changing the paths
			{	wxString filename;
				if ( filepaths.Item(n).Last() != wxFILE_SEP_PATH ) filename = filepaths.Item(n).AfterLast(wxFILE_SEP_PATH);
					else  filename = (filepaths.Item(n).BeforeLast(wxFILE_SEP_PATH)).AfterLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH; // If it's a dir, manoevre round the terminal '/'
				newpaths.Add( destpath + filename );
			}

		if ( ! destctrl->arcman->GetArc()->Alter( filepaths, Moving ? arc_rename : arc_dup, newpaths ) )	// Do the rename. NB we use the destination pane here, and (maybe) reload the origin
				return false;
		
		if ( originctrl != destctrl && originctrl != destctrl->partner )							// If we have 2 editions of the  same archive open, reload the origin one
			{	archivename = ffs->GetRootDir()->GetFilepath(); archivename = archivename.BeforeLast(wxFILE_SEP_PATH);
				originctrl->arcman->GetArc()->Getffs()->Clear();								// Throw out the old dir structure, and reload
				if ( ! IsNested() )
					{ if ( ! archivename.IsEmpty() )  originctrl->arcman->GetArc()->LoadToBuffer( archivename ); }
				 else
					originctrl->arcman->GetArc()->LoadToBuffer( destctrl->arcman->GetArc()->GetBuffer() );	// If we're in a nested buffer situation, copy the updated buffer into the out-of-date one
				originctrl->arcman->GetArc()->ListContentsFromBuffer( archivename );
			}
		
		if ( Moving ) UnRedoManager::StartCluster();
		wxArrayInt IDs; IDs.Add( originctrl->GetId() );
		bool IdenticalArc = (originctrl->arcman->GetArc()->Getffs() == destctrl->arcman->GetArc()->Getffs()); // Are we moving/pasting to a subdir, within the same pane?
		UnRedoArchiveRen* UnRedoptr  = new UnRedoArchiveRen( filepaths, newpaths, IDs, destctrl, 
																					Moving==false, IdenticalArc ? NULL : originctrl ); // Moving==false == not duplicating. Passing destctrl says 'Only if still visible'
		UnRedoManager::AddEntry( UnRedoptr );

		if ( Moving ) 
			{	UnRedoManager::EndCluster();
				MyFrame::mainframe->OnUpdateTrees( OriginPath, IDs );
			}
		MyFrame::mainframe->OnUpdateTrees( destpath, IDs );
		return true;	
	}

 else															//	Otherwise we're going between different archives
	{	wxArrayString tempfiles = originctrl->arcman->ExtractToTempfile( filepaths );	// This extracts the selected filepaths to temporary files, which it returns
		if ( tempfiles.IsEmpty() ) return false;
		wxArrayString destinations;		wxString dest( destpath );
		FakeFiledata* ffd = destctrl->arcman->GetArc()->ffs->GetRootDir()->FindFileByName( dest );		// Destpath doesn't have a '/', so we need to see if it's a file and amputate
		if ( ffd != NULL ) dest = dest.BeforeLast(wxFILE_SEP_PATH);
		if ( dest.Last() != wxFILE_SEP_PATH ) dest << wxFILE_SEP_PATH;
		
		wxArrayString Path; Path.Add( dest ); 														// We have to provide a 2nd arraystring anyway due to the Alter API, so use it to transport the path
		if ( ! destctrl->arcman->GetArc()->Alter( tempfiles, arc_add, Path ) ) return false;	// Do the Paste bit of the Move

		for ( size_t n=0; n < filepaths.GetCount(); ++n )										// Fill destinations array ready for UnRedoing
			{	wxString distal; if ( ! filepaths[n].StartsWith( OriginPath, &distal ) )	continue;
				destinations.Add( dest + distal );
			}

		if ( Moving )																									// If we're Moving (cf Pasting), delete those files from the archive
			{	wxArrayString unused; originctrl->arcman->GetArc()->Alter( filepaths, arc_remove, unused );
				MyFrame::mainframe->OnUpdateTrees( OriginPath, IDs );
				
				UnRedoManager::StartCluster();
				UnRedoArchiveDelete* UnRedoptr = new UnRedoArchiveDelete( filepaths, tempfiles, OriginPath, IDs, originctrl, _("Move")  );
				UnRedoManager::AddEntry( UnRedoptr );
			}

		UnRedoArchivePaste* UnRedoptr = new UnRedoArchivePaste( tempfiles, destinations, dest, IDs, destctrl, Moving ? _("Move") : _("Paste") );
		UnRedoManager::AddEntry( UnRedoptr );
		if ( Moving )		UnRedoManager::EndCluster();
		MyFrame::mainframe->OnUpdateTrees( destpath, IDs );
		return true;
	}
}

wxDEFINE_SCOPED_PTR( wxInputStream, wxInputStreamPtr )
wxDEFINE_SCOPED_PTR( wxOutputStream, wxOutputStreamPtr )
wxDEFINE_SCOPED_PTR_TYPE(wxTarEntry);
wxDEFINE_SCOPED_PTR_TYPE(wxZipEntry);

ZipArchiveStream::ZipArchiveStream( DataBase* archive )	:	ArchiveStream(archive)	// Opens a 'real' archive, not a nested one
{
Valid = false;
if ( ! archive->IsValid() ) return;

wxBusyCursor busy;
if ( ! LoadToBuffer( archive->GetFilepath() ) ) return;	// Load the wxMemoryBuffer with the archive

Valid = true;
}

ZipArchiveStream::ZipArchiveStream( wxMemoryBuffer* membuf, wxString archivename )	:	ArchiveStream(archivename)	// Opens a 'nested' archive
{
Valid = false;
if ( archivename.IsEmpty() ) return;

m_membuf = membuf;
Valid = true;
}

bool ZipArchiveStream::Extract( wxArrayString& filepaths, wxString destpath, wxArrayString& resultingfilepaths )	 // Extract filepath from the archive
{
if ( filepaths.IsEmpty() || destpath.IsEmpty() ) return false;
	// AdjustFilepaths puts each filepath into WithinArchiveNames, less the /path-to-archive/archive; & creates its dest in WithinArchiveNewNames. If a dir, recursively adds into children too
if ( ! AdjustFilepaths( filepaths, destpath, resultingfilepaths, afp_extract ) ) return false;

wxZipEntryPtr entry;
wxMemoryInputStream meminstream( m_membuf->GetData(), m_membuf->GetDataLen() );	// The archive is in the memory buffer m_membuf. 'Extract' it to a memory stream
wxZipInputStream zip( meminstream );

size_t NumberFound = 0;

while (entry.reset(zip.GetNextEntry()), entry.get() != NULL)
	{	int item = WithinArchiveNames.Index( entry->GetName() );	// See if this file is one of the ones we're interested in
		if (  item == wxNOT_FOUND )  continue;
								// Found it. If it's a file, extract it. If a dir,  make a new dir with this name (*outside* the archive ;)
		wxLogNull log;
		if ( ! entry->IsDir() )																	// (For a dir there's no data to extract )
			{	wxString Path(resultingfilepaths[ item ]); Path = Path.BeforeLast(wxFILE_SEP_PATH);
				if ( ! wxFileName::Mkdir( Path, 0777, wxPATH_MKDIR_FULL )) return false;		// Make any necessary dir

				wxFFileOutputStream out( resultingfilepaths[ item ] );
				if ( ! out || ! out.Write( zip ) || ! zip.Eof())   return false;	// and extract into the new filepath
				FileData fd( resultingfilepaths[ item ] ); if ( fd.IsValid() ) fd.DoChmod( entry->GetMode() );	// Correct the extracted file's permissions
			}
		 else if ( ! wxFileName::Mkdir( resultingfilepaths[ item ], 0777, wxPATH_MKDIR_FULL )) return false;	// Just create the dir
		 
		 if ( ++NumberFound >= WithinArchiveNames.GetCount() ) return true;	// See if we've used all the passed filepaths
	}

return false;
}

bool ZipArchiveStream::DoAlteration( wxArrayString& filepaths, enum alterarc DoWhich )	// Implement add remove del or rename, depending on the enum
{
wxInputStreamPtr dupptr;
GetFromBuffer();																	// The archive is in the memory buffer m_membuf. 'Extract' it to InStreamPtr
if ( DoWhich==arc_dup )													// If we're duplicating, we need a 2nd instream, so store the first and re-extract
	{	dupptr.reset( InStreamPtr.release() ); GetFromBuffer( true ); }

wxMemoryOutputStream* memoutstream = new wxMemoryOutputStream;
wxZipOutputStream outzip(*memoutstream);

outzip.CopyArchiveMetaData( *(wxZipInputStream*)InStreamPtr.get() );

wxZipEntryPtr entry,  dupentry;
while ( entry.reset( ((wxZipInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )			// Go thru the archive, looking for the selected files.
	{	int item = WithinArchiveNames.Index( entry->GetName() );	// See if this file is one of the ones we're interested in
		switch( DoWhich )
			{	case arc_add:			break;					// Add happens later (and there shouldn't be a match anyway)
				case arc_remove:	if (  item != wxNOT_FOUND )  continue;	// If this file is one of the 2b-removed ones, ignore it, don't add it to the new archive
													else break;												// Otherwise break to copy the entry
				case arc_dup:			dupentry.reset( ((wxZipInputStream*)dupptr.get())->GetNextEntry() );	// For dup, get the next entry, then fall thru to rename
				case arc_rename:	if (  item != wxNOT_FOUND ) 					// For rename, change the name then break to copy the entry
													{	if ( entry->IsDir() && WithinArchiveNewNames[ item ].Last() != wxFILE_SEP_PATH ) WithinArchiveNewNames[ item ] << wxFILE_SEP_PATH;
														entry->SetName( WithinArchiveNewNames[ item ] ); 
													} 
												break;
													
				case arc_del:			continue;
			}
								// Copy this entry to the outstream
		if ( ! entry->IsDir() )
			{	if ( ! outzip.CopyEntry( entry.release(), *(wxZipInputStream*)InStreamPtr.get() ) ) return false; 	// CopyEntry is specific to archive/zip-outputsteams, which is why I'm using one here
				if ( DoWhich==arc_dup && item != wxNOT_FOUND ) 				// If we're duplicating, we now need to output the original version to the stream
						if ( ! outzip.CopyEntry( dupentry.release(), *(wxZipInputStream*)dupptr.get() ) ) return false;
			}
		 else 
			{	if ( ! outzip.PutNextDirEntry( entry->GetName(), entry->GetDateTime() ) ) return false; 	// If it's a dir, this just writes the 'header' info; there isn't any data
				entry.release();
				if ( DoWhich==arc_dup && item != wxNOT_FOUND )  				// If we're duplicating, we now need to output the original version to the stream
					{ if ( ! outzip.PutNextDirEntry( dupentry->GetName(), dupentry->GetDateTime() ) ) return false; dupentry.release(); }
			}
	}	

if ( DoWhich == arc_add )		// If we're trying to add, we've just copied all the old entries. Now add new the files to the within-archive dir WithinArchiveName
	{	bool success = false;
		wxString path =  filepaths[0].BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;	// The first item *must* exist. Use it to find the curent path of the entries: which will be amputated
		for ( size_t n=0; n < filepaths.GetCount(); ++n )
			{	FileData fd( filepaths[n] ); if ( ! fd.IsValid() ) continue;
				wxString Name; if ( ! filepaths[n].StartsWith( path, &Name ) )	continue;			// Get what's after path into Name: if we've adding  dir foo, path==../path/, Name will be /foo or /foo/bar
				Name = WithinArchiveName + Name; wxDateTime DT( fd.ModificationTime() );
				if ( fd.IsDir() )
					{ success = outzip.PutNextDirEntry( Name, DT ); continue; }	// Dirs don't have data, so just provide the filepath

				wxFFileInputStream file( filepaths[n] ); if ( ! file.Ok() ) continue;	// Make a wxFFileInputStream to get the data to be added
				wxZipEntry* addentry = new wxZipEntry( Name, DT, file.GetSize() ); addentry->SetSystemMadeBy( wxZIP_SYSTEM_UNIX ); addentry->SetMode( fd.GetPermissions() );
				if ( ! outzip.PutNextEntry( addentry ) 						// and add it to the outputstream
												|| ! outzip.Write( file ) || ! file.Eof() )    continue;
				success = true;																				// Flag that we've at least one successful addition
			}
		if ( ! success ) return false;																	// No point proceeding if nothing worked
	}

if ( ! outzip.Close() ) return false;
if ( ! memoutstream->Close() ) return false;

OutStreamPtr.reset( memoutstream );								// This has a kludgey feel, but it returns the memoutstream to Alter() via the member scoped ptr
return true;
}

wxMemoryBuffer* ZipArchiveStream::ExtractToBuffer( wxString filepath )	 // Extract filepath from this archive into a new wxMemoryBuffer
{
if ( filepath.IsEmpty() ) return NULL;

if ( ! SortOutNames( filepath ) ) return NULL;

wxZipEntryPtr entry;

wxBusyCursor busy;
wxMemoryInputStream meminstream( m_membuf->GetData(), m_membuf->GetDataLen() );	// The archive is in the memory buffer m_membuf. 'Extract' it to a memory stream
wxZipInputStream zip( meminstream );

while (entry.reset(zip.GetNextEntry()), entry.get() != NULL)
	{	if ( entry->GetName() != WithinArchiveName )  continue;											// Go thru the archive, looking for the desired file
								// Found it. Extract it to a new buffer
		wxMemoryOutputStream memoutstream;
		if ( ! memoutstream || ! memoutstream.Write( zip ) )   return NULL;	

		size_t size = memoutstream.GetSize();																		// This is how much was loaded i.e. the file size
		wxChar* buf = new wxChar[ size ];																			// It's ugly, but the easiest way I've found to get the data from the stream to a memorybuffer
		memoutstream.CopyTo( buf, size );																			//   is by using an intermediate char[]

		wxMemoryBuffer* membuf = new wxMemoryBuffer( size );										// Now (at last) we can store the data in a memory buffer
		membuf->AppendData( buf, size );
		delete[] buf;
	
		return membuf;
	}

return NULL;
}

void ZipArchiveStream::ListContentsFromBuffer( wxString archivename )
{
wxBusyCursor busy;
GetFromBuffer();
ListContents( archivename );
}

void ZipArchiveStream::GetFromBuffer( bool dup )		// Get the stored archive from membuf into the stream
{
wxZipInputStream* zip;
wxMemoryInputStream* meminstream = new wxMemoryInputStream( m_membuf->GetData(), m_membuf->GetDataLen() );	// The archive is in the memory buffer membuf. Stream it to a memory stream
if ( ! dup ) 
	{	MemInStreamPtr.reset( meminstream );
 		zip = new wxZipInputStream( *MemInStreamPtr.get() );					// Then pass it thru the filter
	}
 else
	{	DupMemInStreamPtr.reset( meminstream );
 		zip = new wxZipInputStream( *DupMemInStreamPtr.get() );				// If we're duplicating, use DupMemInStreamPtr the second time around
	}

InStreamPtr.reset( zip );																				// Use this member wxScopedPtr to 'return' the stream
}

void ZipArchiveStream::ListContents( wxString archivename )
{
				// Zip files don't contain absolute paths, just any subdirs, so get the bit to be prepended
if ( archivename.GetChar( archivename.Len()-1) != wxFILE_SEP_PATH ) archivename << wxFILE_SEP_PATH;

wxZipEntryPtr entry;

while (entry.reset( ((wxZipInputStream*)InStreamPtr.get())->GetNextEntry()), entry.get() != NULL)
	{	 entry->SetSystemMadeBy( wxZIP_SYSTEM_UNIX );
		wxString name = archivename + entry->GetName();
		wxDateTime time = entry->GetDateTime();
		// read 'zip' to access the entry's data
		ffs->AddItem( name, entry->GetSize(), time.GetTicks(), entry->GetMode(), entry->IsDir() ? DIRTYPE : REGTYPE ); // wxZipEntry can't cope with symlinks etc 
	}
}

void ZipArchiveStream::RefreshFromDirtyChild( wxString archivename, ArchiveStream* child )	// Used when a nested child archive has been altered; reload the altered version
{
if ( archivename.IsEmpty()  || child == NULL ) return;
wxMemoryInputStream  childstream( child->GetBuffer()->GetData(), child->GetBuffer()->GetDataLen() );

GetFromBuffer();																	// The archive is in the memory buffer m_membuf. 'Extract' it to InStreamPtr
wxMemoryOutputStream memoutstream;
wxZipOutputStream outzip(memoutstream);

outzip.CopyArchiveMetaData( *(wxZipInputStream*)InStreamPtr.get() );

wxZipEntryPtr entry;
while ( entry.reset( ((wxZipInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )		// Go thru the archive, looking for the selected files.
	{									// Copy this entry to the outstream
		if ( ! entry->IsDir() )
			{	if ( archivename == entry->GetName() )  			// If this is the archive we want to refresh
					{ if ( ! outzip.PutNextEntry( entry->GetName(), entry->GetDateTime(), childstream.GetSize() ) 	// ignore the original archive; instead add the new data
												|| ! outzip.Write( childstream ) || ! childstream.Eof() )    return;
					}
				 else if ( ! outzip.CopyEntry( entry.release(), *(wxZipInputStream*)InStreamPtr.get() ) ) return; 	// CopyEntry is specific to archive/zip-outputsteams, which is why I'm using one here

			}
		 else 
			{	if ( ! outzip.PutNextDirEntry( entry->GetName(), entry->GetDateTime() ) ) return; 	// If it's a dir, this just writes the 'header' info; there isn't any data
				entry.release();
			}
	}	

if ( ! outzip.Close() )  return;
if ( ! memoutstream.Close() )  return;

size_t size = memoutstream.GetSize();								// This is how much was loaded i.e. the file size
wxChar* buf = new wxChar[ size ];									// It's ugly, but the easiest way I've found to get the data from the stream to a memorybuffer
memoutstream.CopyTo( buf, size );									// is by using an intermediate char[]

delete m_membuf; m_membuf = new wxMemoryBuffer( size ); // I've no idea why, but trying to delete the old data by SetDataLen( 0 ) failed :(
m_membuf->AppendData( buf, size );									// Store the data
delete[] buf;

SaveBuffer();																		// We now need to save (or declare dirty) the parent arc
}

//-----------------------------------------------------------------------------------------------------------------------

TarArchiveStream::TarArchiveStream( DataBase* archive )	:	ArchiveStream(archive)	// Opens a 'real' archive, not a nested one
{
Valid = false;
if ( ! archive->IsValid() ) return;

wxBusyCursor busy;
if ( ! LoadToBuffer( archive->GetFilepath() ) ) return;	// Common code, so do this in base class. Loads the wxMemoryBuffer with the archive
Valid = true;
}

TarArchiveStream::TarArchiveStream( wxMemoryBuffer* membuf, wxString archivename )	:	ArchiveStream(archivename)	// Opens a 'nested' archive
{
Valid = false;
if ( archivename.IsEmpty() ) return;

m_membuf = membuf;
Valid = true;
}

void TarArchiveStream::ListContentsFromBuffer( wxString archivename )
{
wxBusyCursor busy;
GetFromBuffer();
ListContents( archivename );
}

void TarArchiveStream::ListContents( wxString archivename )
{
				// Tar files don't contain absolute paths, just any subdirs, so get the bit to be prepended
if ( archivename.GetChar( archivename.Len()-1) != wxFILE_SEP_PATH ) archivename << wxFILE_SEP_PATH;

wxTarEntryPtr entry;
while ( entry.reset( ((wxTarInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )
	{	wxString target;
		wxString name = archivename + entry->GetName();
		wxDateTime time = entry->GetDateTime();
		DB_filetype Type;	// wxTarEntry (cf. zipentry) stores the type of 'file' e.g. symlink. So make use of that
		switch( entry->GetTypeFlag() )
			{	case wxTAR_DIRTYPE: Type = DIRTYPE; break; case wxTAR_FIFOTYPE: Type = FIFOTYPE; break;
				case wxTAR_BLKTYPE: Type = BLKTYPE; break; case wxTAR_CHRTYPE: Type = CHRTYPE; break;
				case wxTAR_SYMTYPE: Type = SYMTYPE; target = entry->GetLinkName(); break; 
				 default: Type = REGTYPE;	// Not only reg files, but hardlinks, ex-sockets & "contiguous files" (see wxTAR_CONTTYPE and http://www.gnu.org/software/tar/manual/html_node/Standard.html)
			}

		ffs->AddItem( name, entry->GetSize(), time.GetTicks(),  entry->GetMode(), Type, target );
	}
}

void TarArchiveStream::GetFromBuffer( bool dup )		// Get the stored archive from m_membuf into the stream
{
wxMemoryInputStream* meminstream = new wxMemoryInputStream( m_membuf->GetData(), m_membuf->GetDataLen() );	// The archive is in the memory buffer membuf. Stream it to a memory stream
wxTarInputStream* tar;
if ( ! dup ) 
	{	MemInStreamPtr.reset( meminstream );
		tar = new wxTarInputStream( *MemInStreamPtr.get() );
	}
 else
 	{	DupMemInStreamPtr.reset( meminstream );										// If we're duplicating, use DupMemInStreamPtr the second time around
		tar = new wxTarInputStream( *DupMemInStreamPtr.get() );
	}

InStreamPtr.reset( tar );																				// Use this member wxScopedPtr to 'return' the stream
}

bool TarArchiveStream::CompressStream( wxOutputStream* outstream )		// Pretend to compress this stream, ready to be put into membuf
{
OutStreamPtr.reset( outstream );	// Since we can't compress a plain vanilla tarstream, use the member wxScopedPtr to 'return' the same stream
return true;
}

bool TarArchiveStream::Extract( wxArrayString& filepaths, wxString destpath, wxArrayString& resultingfilepaths )	 // Create and extract filepath from the archive
{
if ( filepaths.IsEmpty() || destpath.IsEmpty() ) return false;

	// AdjustFilepaths puts each filepath into WithinArchiveNames, less the /path-to-archive/archive; & creates its dest in WithinArchiveNewNames. If a dir, recursively adds into children too
if ( ! AdjustFilepaths( filepaths, destpath, resultingfilepaths, afp_extract ) ) return false;

GetFromBuffer();

size_t NumberFound = 0;
wxTarEntryPtr entry;

while ( entry.reset( ((wxTarInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )
	{	int item = WithinArchiveNames.Index( entry->GetName() );	// See if this file is one of the ones we're interested in
		if (  item == wxNOT_FOUND )  continue;
								// Found it. If it's a file, extract it. If a dir,  make a new dir with this name (*outside* the archive ;)
		wxLogNull log;
		if ( entry->GetTypeFlag() == wxTAR_REGTYPE )					// Start with the files, which are the only things that require extraction of data
			{	wxString Path(resultingfilepaths[ item ]); Path = Path.BeforeLast(wxFILE_SEP_PATH);
				if ( ! wxFileName::Mkdir( Path, 0777, wxPATH_MKDIR_FULL )) return false;		// Make any necessary dir
				wxFFileOutputStream out( resultingfilepaths[ item ] );
				if ( ! out || ! out.Write( *(wxTarInputStream*)InStreamPtr.get() ) || ! InStreamPtr->Eof() )   return false;	// and extract into the new filepath
				FileData fd( resultingfilepaths[ item ] ); if ( fd.IsValid() ) fd.DoChmod( entry->GetMode() );	// Correct the extracted file's permissions
			}
		 else if ( entry->IsDir() )	
		 			{ if ( ! wxFileName::Mkdir( resultingfilepaths[ item ], 0777, wxPATH_MKDIR_FULL )) return false; }	// Just create the dir

		else if ( entry->GetTypeFlag() == wxTAR_SYMTYPE || entry->GetTypeFlag() == wxTAR_LNKTYPE ) // The 2nd is in case of confusion, but all links should be symlinks
		 			{ if ( ! CreateSymlink( entry->GetLinkName(), resultingfilepaths[ item ] ) ) return false; }
					
		else				// Otherwise it's a Misc:  a FIFO or similar. Just create a new one with the same name
			{	wxString Path(resultingfilepaths[ item ]); Path = Path.BeforeLast(wxFILE_SEP_PATH);
				FileData fd( Path, true );
				 mode_t mode; dev_t dev=0;
				switch( entry->GetTypeFlag() )
					{	case wxTAR_CHRTYPE:	mode = S_IFCHR; dev = fd.GetDeviceID(); break; case wxTAR_BLKTYPE: mode = S_IFBLK; dev = fd.GetDeviceID(); break;
						case wxTAR_FIFOTYPE: mode = S_IFIFO; break; 					// FIFOs don't need the major/minor thing in dev
						 default: continue; 
					}

				mode_t oldUmask = umask(0); 
				if ( entry->GetTypeFlag()==wxTAR_FIFOTYPE || getuid()==0 )		// If we're making a fifo (which doesn't require su), or we're already root
					mknod( resultingfilepaths[ item ].mb_str(wxConvUTF8), mode | 0777, dev );
				 else
					{	wxString secondbit = (mode==S_IFBLK ? wxT(" b ") :  wxT(" c ")); secondbit << (int)(dev / 256) << wxT(" ") << (int)(dev % 256);
						if ( WHICH_SU==kdesu )  wxExecute( wxT("kdesu \"mknod -m 0777 ") + resultingfilepaths[ item ] + secondbit + wxT("\"") );
						 else if ( WHICH_SU==gksu )  wxExecute( wxT("gksu \"mknod -m 0777 ") + resultingfilepaths[ item ] + secondbit + wxT("\"") );
						 else if ( WHICH_SU==gnomesu )  wxExecute( wxT("gnomesu --title "" -c \"mknod -m 0777 ") + resultingfilepaths[ item ] + secondbit + wxT("\" -d") );
							  else { wxMessageBox(_("Sorry, you need to be root to extract character or block devices"), wxT(":-(") ); umask( oldUmask ); return false; }
					}
				umask( oldUmask );
			}
		 
		 if ( ++NumberFound >= WithinArchiveNames.GetCount() ) return true;	// See if we've used all the passed filepaths
	}

return false;
}

wxMemoryBuffer* TarArchiveStream::ExtractToBuffer( wxString filepath )	 // Extract filepath from this archive into a new wxMemoryBuffer
{
if ( filepath.IsEmpty() ) return NULL;

if ( ! SortOutNames( filepath ) ) return NULL;

wxTarEntryPtr entry;

wxBusyCursor busy;
GetFromBuffer();					// The archive is in the memory buffer m_membuf. 'Extract' it to InStreamPtr

while ( entry.reset( ((wxTarInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )
	{	if ( entry->GetName() != WithinArchiveName )  continue;											// Go thru the archive, looking for the desired file
								// Found it. Extract it to a new buffer
		wxMemoryOutputStream memoutstream;
		if ( ! memoutstream || ! memoutstream.Write( *(wxTarInputStream*)InStreamPtr.get() ) )   return NULL;	

		size_t size = memoutstream.GetSize();																		// This is how much was loaded i.e. the file size
		wxChar* buf = new wxChar[ size ];																			// It's ugly, but the easiest way I've found to get the data from the stream to a memorybuffer
		memoutstream.CopyTo( buf, size );																			//   is by using an intermediate char[]

		wxMemoryBuffer* membuf = new wxMemoryBuffer( size );										// Now (at last) we can store the data in a memory buffer
		membuf->AppendData( buf, size );
		delete[] buf;
	
		return membuf;
	}

return NULL;
}

bool TarArchiveStream::DoAlteration( wxArrayString& filepaths, enum alterarc DoWhich )		// Implement add remove del or rename, depending on the enum
{
wxInputStreamPtr dupptr;
GetFromBuffer();																	// The archive is in the memory buffer m_membuf. 'Extract' it to InStreamPtr
if ( DoWhich==arc_dup )													// If we're duplicating, we need a 2nd instream, so store the first and re-extract
	{	dupptr.reset( InStreamPtr.release() ); GetFromBuffer( true ); }

wxMemoryOutputStream* memoutstream = new wxMemoryOutputStream;
if ( ! CompressStream( memoutstream ) ) return false;		// This will compress the output (unless we're dealing with a plain tar). False means we're not gzip-capable

wxTarOutputStream taroutstream( *OutStreamPtr.get() );	// OutStreamPtr is the 'output' of the CompressStream call in Alter()

wxTarEntryPtr entry,  dupentry;
while ( entry.reset( ((wxTarInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )		// Go thru the archive, looking for the selected files.
	{	int item = WithinArchiveNames.Index( entry->GetName() );	// See if this file is one of the ones we're interested in
		switch( DoWhich )
			{	case arc_add:			break;					// Add happens later (and there shouldn't be a match anyway)
				case arc_remove:	if (  item != wxNOT_FOUND )  continue;	// If this file is one of the 2b-removed ones, ignore it, don't add it to the new archive
													else break;												// Otherwise break to copy the entry
				case arc_dup:			dupentry.reset( ((wxTarInputStream*)dupptr.get())->GetNextEntry() );	// For dup, get the next entry, then fall thru to rename
				case arc_rename:	if (  item != wxNOT_FOUND ) 					// For rename, change the name then break to copy the entry
													{	if ( entry->IsDir() && WithinArchiveNewNames[ item ].Last() != wxFILE_SEP_PATH ) WithinArchiveNewNames[ item ] << wxFILE_SEP_PATH;
														entry->SetName( WithinArchiveNewNames[ item ] ); 
													} 
												break;
													
				case arc_del:			continue;
			}
								// Copy this entry to the outstream
		if ( ! entry->IsDir() )
			{	if ( ! taroutstream.CopyEntry( entry.release(), *(wxTarInputStream*)InStreamPtr.get() ) ) return false; 	// CopyEntry is specific to archive/zip-outputsteams, which is why I'm using one here
				if ( DoWhich==arc_dup && item != wxNOT_FOUND ) 				// If we're duplicating, we now need to output the original version to the stream
						if ( ! taroutstream.CopyEntry( dupentry.release(), *(wxTarInputStream*)dupptr.get() ) ) return false;
			}
		 else 
			{	if ( ! taroutstream.PutNextDirEntry( entry->GetName(), entry->GetDateTime() ) ) return false; 	// If it's a dir, this just writes the 'header' info; there isn't any data
				entry.release();
				if ( DoWhich==arc_dup && item != wxNOT_FOUND )  				// If we're duplicating, we now need to output the original version to the stream
					{ if ( ! taroutstream.PutNextDirEntry( dupentry->GetName(), dupentry->GetDateTime() ) ) return false; dupentry.release(); }
			}
	}	

if ( DoWhich == arc_add )		// If we're trying to add, we've just copied all the old entries. Now add new the files to the within-archive dir WithinArchiveName
	{	bool success = false;
		wxString path =  filepaths[0].BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH;	// The first item *must* exist. Use it to find the curent path of the entries: which will be amputated
		for ( size_t n=0; n < filepaths.GetCount(); ++n )
			{	FileData fd( filepaths[n] ); if ( ! fd.IsValid() ) continue;
				wxString Name; if ( ! filepaths[n].StartsWith( path, &Name ) )	continue;	// Get what's after path into Name: if we've adding  dir foo, path==../path/, Name will be /foo or /foo/bar
				Name = WithinArchiveName + Name; wxDateTime DT( fd.ModificationTime() );
				if ( fd.IsDir() )
					{ if ( taroutstream.PutNextDirEntry( Name, DT ) ) success = true; continue; }		// Dirs don't have data, so just provide the filepath

				wxTarEntry* addentry = new wxTarEntry( Name, DT ); addentry->SetMode( fd.GetPermissions() );
				if ( fd.IsRegularFile() )
					{	wxFFileInputStream file( filepaths[n] ); if ( ! file.Ok() ) continue;	// Make a wxFFileInputStream to get the data to be added
						addentry->SetSize( file.GetSize() );
						if ( ! taroutstream.PutNextEntry( addentry )								 	// and add it to the outputstream
														|| ! taroutstream.Write( file ) || ! file.Eof() )    continue;
						success = true;	continue;												// Flag that we've at least one successful addition
					}
														// If it's not a dir, or a file, it must be a misc
				int type = wxTAR_REGTYPE; if ( fd.IsSymlink() ) type = wxTAR_SYMTYPE;
				if ( fd.IsCharDev() ) type = wxTAR_CHRTYPE; if ( fd.IsBlkDev() ) type = wxTAR_BLKTYPE; if ( fd.IsFIFO() ) type = wxTAR_FIFOTYPE;
				addentry->SetTypeFlag( type );
				if ( fd.IsSymlink() ) addentry->SetLinkName( fd.GetSymlinkDestination(true) );	// If a symlink, tell entry its target
				if ( ! taroutstream.PutNextEntry( addentry ) )    continue;			// Add it to the outputstream
				
				success = true;
			}
		if ( ! success ) return false;																// No point proceeding if nothing worked
	}

if ( ! taroutstream.Close() ) return false;
if ( ! OutStreamPtr.get()->Close() )  return false;
if ( ! memoutstream->Close() ) return false;

OutStreamPtr.reset( memoutstream );								// This has a kludgey feel, but it returns the memoutstream to Alter() via the member scoped ptr
return true;
}

void TarArchiveStream::RefreshFromDirtyChild( wxString archivename, ArchiveStream* child )	// Used when a nested child archive has been altered; reload the altered version
{
if ( archivename.IsEmpty()  || child == NULL ) return;
wxMemoryInputStream  childstream( child->GetBuffer()->GetData(), child->GetBuffer()->GetDataLen() );

GetFromBuffer();																	// The archive is in the memory buffer m_membuf. 'Extract' it to InStreamPtr
wxMemoryOutputStream* memoutstream = new wxMemoryOutputStream;
if ( ! CompressStream( memoutstream ) ) return;				// This will compress the output (unless we're dealing with a plain tar). False means we're not gzip-capable
wxTarOutputStream taroutstream( *OutStreamPtr.get() );	// OutStreamPtr is the 'output' of the CompressStream call in Alter()

wxTarEntryPtr entry;
while ( entry.reset( ((wxTarInputStream*)InStreamPtr.get())->GetNextEntry() ), entry.get() != NULL )		// Go thru the archive, looking for the selected files.
	{									// Copy this entry to the outstream
		if ( ! entry->IsDir() )
			{	if ( archivename == entry->GetName() )  			// If this is the archive we want to refresh
					{ if ( ! taroutstream.PutNextEntry( entry->GetName(), entry->GetDateTime(), childstream.GetSize() ) 	// ignore the original archive; instead add the new data
												|| ! taroutstream.Write( childstream ) || ! childstream.Eof() )    return;
					}
				 else if ( ! taroutstream.CopyEntry( entry.release(), *(wxTarInputStream*)InStreamPtr.get() ) ) return; 	// CopyEntry is specific to archive/zip-outputsteams, which is why I'm using one here

			}
		 else 
			{	if ( ! taroutstream.PutNextDirEntry( entry->GetName(), entry->GetDateTime() ) ) return; 	// If it's a dir, this just writes the 'header' info; there isn't any data
				entry.release();
			}
	}	

if ( ! taroutstream.Close() ) return;
if ( ! OutStreamPtr->Close() ) return;
if ( ! memoutstream->Close() )return;

size_t size = memoutstream->GetSize();							// This is how much was loaded i.e. the file size
wxChar* buf = new wxChar[ size ];									// It's ugly, but the easiest way I've found to get the data from the stream to a memorybuffer
memoutstream->CopyTo( buf, size );									// is by using an intermediate char[]
delete memoutstream;

delete m_membuf; m_membuf = new wxMemoryBuffer( size ); // I've no idea why, but trying to delete the old data by SetDataLen( 0 ) failed :(
m_membuf->AppendData( buf, size );									// Store the data
delete[] buf;

SaveBuffer();																		// We now need to save (or declare dirty) the parent arc
}

//-----------------------------------------------------------------------------------------------------------------------

void GZArchiveStream::GetFromBuffer( bool dup )		// Get the stored archive from membuf into the stream
{
wxZlibInputStream* zlib;  wxTarInputStream* tar;
wxMemoryInputStream* meminstream = new wxMemoryInputStream( m_membuf->GetData(), m_membuf->GetDataLen() );	// The archive is in the memory buffer membuf. Stream it to a memory stream
if ( ! dup ) 
	{	MemInStreamPtr.reset( meminstream );
		zlib = new wxZlibInputStream( *MemInStreamPtr.get() );											// Then pass it thru the filter
		zlibInStreamPtr.reset( zlib );
		tar = new wxTarInputStream( *zlibInStreamPtr.get() );												// into a wxTarInputStream
	}
 else
 	{	DupMemInStreamPtr.reset( meminstream );
		zlib = new wxZlibInputStream( *DupMemInStreamPtr.get() );									// If we're duplicating, use DupMemInStreamPtr/DupzlibInStreamPtr the second time around
		DupzlibInStreamPtr.reset( zlib );
		tar = new wxTarInputStream( *DupzlibInStreamPtr.get() );
	}

InStreamPtr.reset( tar );																										// Use this member wxScopedPtr to 'return' the stream
}

bool GZArchiveStream::CompressStream( wxOutputStream* outstream )		// Compress this stream, ready to be put into membuf
{
if ( ! wxZlibOutputStream::CanHandleGZip() )
	{ wxLogError(_("I'm afraid your zlib is too old to be able to do this :(") ); return false; }

wxZlibOutputStream* zlib = new wxZlibOutputStream( *outstream, wxZ_BEST_COMPRESSION, wxZLIB_GZIP );	// Pass the outstream thru the filter
OutStreamPtr.reset( zlib );																																			// Use the member wxScopedPtr to 'return' the stream

return true;
}

void GZArchiveStream::ListContentsFromBuffer( wxString archivename )
{
wxBusyCursor busy;
GetFromBuffer();
ListContents( archivename );
}

//-----------------------------------------------------------------------------------------------------------------------
void BZArchiveStream::GetFromBuffer( bool dup )		// Get the stored archive from membuf into the stream
{
wxBZipInputStream* zlib; wxTarInputStream* tar;
wxMemoryInputStream* meminstream = new wxMemoryInputStream( m_membuf->GetData(), m_membuf->GetDataLen() );	// The archive is in the memory buffer membuf. Stream it to a memory stream
if ( ! dup ) 
	{	MemInStreamPtr.reset( meminstream );
		zlib = new wxBZipInputStream( *MemInStreamPtr.get() );											// Then pass it thru the filter
		zlibInStreamPtr.reset( zlib );
		tar = new wxTarInputStream( *zlibInStreamPtr.get() );												// into a wxTarInputStream
	}
  else
	{	DupMemInStreamPtr.reset( meminstream );
		zlib = new wxBZipInputStream( *DupMemInStreamPtr.get() );									// If we're duplicating, use DupMemInStreamPtr/DupzlibInStreamPtr the second time around
		DupzlibInStreamPtr.reset( zlib );
		tar = new wxTarInputStream( *DupzlibInStreamPtr.get() );
	}

InStreamPtr.reset( tar );																										// Use this member wxScopedPtr to 'return' the stream
}

bool BZArchiveStream::CompressStream( wxOutputStream* outstream )			// Compress this stream, ready to be put into membuf
{
wxBZipOutputStream* bzip = new wxBZipOutputStream( *outstream, 9 );			// Pass the outstream thru the filter
OutStreamPtr.reset( bzip );																					// Use the member wxScopedPtr to 'return' the stream

return true;
}

void BZArchiveStream::ListContentsFromBuffer( wxString archivename )
{
wxBusyCursor busy;
GetFromBuffer();
ListContents( archivename );
}

#endif  // wxVERSION_NUMBER >= 2504
//-----------------------------------------------------------------------------------------------------------------------

void ArchiveStreamMan::Push()	// Make a new ArchiveStruct and save it on the array
{
if ( arc==NULL ) return;

ArchiveStruct* newarcst = new struct ArchiveStruct( arc, ArchiveType );
ArcArray->Add( newarcst );
}

bool ArchiveStreamMan::Pop()		// Revert to any previous archive. Returns true if this happened
{
#if wxVERSION_NUMBER >= 2504
size_t count = ArcArray->GetCount();
if ( ! count  )																												// If we're not within multiple archives, just delete and exit
	{	if ( ! IsArchive() ) return false;																			//  providing there's anything to delete
		delete arc; arc = NULL; ArchiveType = NotArchive; OriginalArchive.Empty(); 	// NB We also kill any arc data, so Pop() will exit a 'single' archive too
						// NB we really need the following checks, as otherwise we'll segfault if we're within an archive when the program exits
		if ( MyFrame::mainframe==NULL || ! MyFrame::mainframe->SelectionAvailable ) return false;
		MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane(); if ( active==NULL ) return false; 
		if ( active->fileview == ISRIGHT ) active = active->partner; active->fulltree = WasOriginallyFulltree;	// Set the pane's fulltree-ness to its original value
		return false; 
	}

ArchiveStruct* PreviousArcst = ArcArray->Item( count-1 );
if ( arc->IsDirty() ) PreviousArcst->arc->RefreshFromDirtyChild( arc->GetArchiveName().AfterLast(wxFILE_SEP_PATH), arc );	// If we're exiting a nested archive, it might need 2b saved to its parent
delete arc; arc = PreviousArcst->arc; ArchiveType = PreviousArcst->ArchiveType;
ArcArray->RemoveAt( count-1 ); delete PreviousArcst;				// NB deleting the struct doesn't delete the stored archive, as it is referenced by arc.

return true;
#else
return false;
#endif
}

bool ArchiveStreamMan::NewArc( wxString filepath, enum ziptype NewArchiveType )
{
  // We *could* use zipstreams from 2.5.3 (and plain tars anytime), but bzip streams need wxOutputStream::Close which starts in 2.5.4
#if wxVERSION_NUMBER < 2504
switch( NewArchiveType )
		{	case zip:case targz:case taronly:case tarbz:	wxMessageBox(_("I'm afraid this isn't possible from this version of wxWidgets. You need to upgrade to at least version 2.5.4.") ); return false;
			 default: return false;
		}
#else
if ( filepath.IsEmpty() ) return false;

DataBase* stat;
bool OriginallyInArchive = IsArchive();
if ( IsArchive() ) stat = new FakeFiledata( filepath );	// If we're opening an archive within an archive, we need to use FakeFiledata here
 else stat = new FileData( filepath );
if ( ! stat->IsValid() ) { delete stat; return false; }

ArchiveStream* newarc;

switch( NewArchiveType )
		{	case htb:																											// htb's are effectively zips
			case zip: 		if ( ! IsArchive() ) newarc = new ZipArchiveStream( stat ); 	// If we're not already inside an archive, do things the standard way
									 else
										{
										wxMemoryBuffer* membuf; membuf = arc->ExtractToBuffer( filepath );
										if ( membuf == NULL ) { delete stat; return false; }
										newarc = new ZipArchiveStream( membuf, filepath );
										}
										 if ( ! ( newarc && newarc->IsValid() ) )
									 	{	wxLogError(_("For some reason, the archive failed to open  :( ") ); 
									 			delete stat; return false;
											}
										((ZipArchiveStream*)newarc)->ListContentsFromBuffer( filepath ); break;
										break;
			case targz:		if ( ! IsArchive() )  newarc = new GZArchiveStream( stat );	// If we're not already inside an archive, do things the standard way
									 else
									 	{	wxMemoryBuffer* membuf; membuf = arc->ExtractToBuffer( filepath );	// Extract filepath from the previous archive to a membuffer
											if ( membuf == NULL ) { delete stat; return false; }
											newarc = new GZArchiveStream( membuf, filepath );	// Make a new archive of it
										}
									 if ( ! ( newarc && newarc->IsValid() ) )
									 	{	wxLogError(_("For some reason, the archive failed to open  :( ") ); 
									 		delete stat; return false;
										}
									 ((GZArchiveStream*)newarc)->ListContentsFromBuffer( filepath ); break;
			case taronly:	if ( ! IsArchive() )  newarc = new TarArchiveStream( stat );
									 else
									 	{	wxMemoryBuffer* membuf; membuf = arc->ExtractToBuffer( filepath );
											if ( membuf == NULL ) { delete stat; return false; }
											newarc = new TarArchiveStream( membuf, filepath );
										}
									 if ( ! ( newarc && newarc->IsValid() ) )
									 	{	wxLogError(_("For some reason, the archive failed to open  :( ") ); 
									 		delete stat; return false;
										}
									 ((TarArchiveStream*)newarc)->ListContentsFromBuffer( filepath ); break;
			case tarbz:		if ( ! IsArchive() )  newarc = new BZArchiveStream( stat );
									 else
									 	{	wxMemoryBuffer* membuf; membuf = arc->ExtractToBuffer( filepath );
											if ( membuf == NULL ) { delete stat; return false; }
											newarc = new BZArchiveStream( membuf, filepath );
										}
									 if ( ! ( newarc && newarc->IsValid() ) )
									 	{	wxLogError(_("For some reason, the archive failed to open  :( ") ); 
									 		delete stat; return false;
										}
									 ((BZArchiveStream*)newarc)->ListContentsFromBuffer( filepath ); break;
			case gzip:
 			case bzip:		wxMessageBox(_("This file is compressed, but it's not an archive so you can't peek inside."), _("Sorry") ); delete stat; return false;
			 default: 			wxMessageBox(_("I'm afraid I don't know about that sort of archive."), _("Sorry") ); delete stat; return false;
		}
if ( newarc == NULL ) { delete stat; return false; }

Push();																		// Save any current archive data
arc = newarc; ArchiveType = NewArchiveType;			//  and install the new values
if ( ! OriginallyInArchive )											// If this is the original archive (ie not nesting), store interesting info
		{	OriginalArchive = filepath; OutOfArchivePath = filepath.BeforeLast(wxFILE_SEP_PATH) + wxFILE_SEP_PATH; }
	
delete stat; return true;
#endif // ! wxVERSION_NUMBER < 2504
}

bool ArchiveStreamMan::IsWithinArchive( wxString filepath )	// Is filepath within the archive system i.e. the original archive itself, or 1 of its contents, or that of a nested archive
{
if ( arc==NULL || ! IsArchive() ) return false;

while ( filepath.Right(1) == wxFILE_SEP_PATH && filepath.Len() > 1 ) filepath.RemoveLast();
return filepath.StartsWith( OriginalArchive );
}

bool ArchiveStreamMan::IsWithinThisArchive( wxString filepath )
{
#if wxVERSION_NUMBER >= 2504
if ( ! IsWithinArchive( filepath ) ) return false;			// It's faster first to check if it's possible

return arc->IsWithin( filepath ); 
#else
return false;
#endif
}

bool ArchiveStreamMan::MightFilepathBeWithinChildArchive( wxString filepath )	// See if filepath is longer than the current rootdir. Assume we've just checked it's not in *this* archive
{
#if wxVERSION_NUMBER >= 2504
if ( ! IsArchive() || filepath.IsEmpty() ) return false;	// If there isn't an archive open, the question is meaningless

return filepath.StartsWith( GetArc()->Getffs()->GetRootDir()->GetFilepath() );	// Shouldn't need to worry about the terminal '/' on rootdir
#else
return false;
#endif
}

wxString ArchiveStreamMan::FindFirstArchiveinFilepath( wxString filepath )	// FIlepath will be a fragment of a path. Find the first thing that looks like an archive
{
wxString filename, segment;
while (true)
	{	if ( filepath.GetChar(0) == wxFILE_SEP_PATH ) filepath = filepath.Mid(1);	// Remove any prepended '/'
		if ( filepath.IsEmpty() ) return filepath;
		filename += wxFILE_SEP_PATH; segment = filepath.BeforeFirst(wxFILE_SEP_PATH);	// Get the first segment of filepath
		 filename += segment; if ( Archive::Categorise( filename ) != NotArchive ) return filename;	// See if it's an archive. If so return it
		filepath = filepath.Mid( segment.Len() );															// Amputate the bit we just checked from filepath & try again
	}

return wxEmptyString;		// We shouldn't reach here
}


enum OCA ArchiveStreamMan::OpenCorrectArchive( wxString& filepath )	// See if filepath is (within) a, possibly nested, archive. If so, open it
{
#if wxVERSION_NUMBER >= 2504
if ( filepath.IsEmpty() ) return OCA_false;
if ( IsWithinThisArchive( filepath ) ) return OCA_true;	// The trivial case!

wxString archivepath;
if ( ! IsArchive() )
	{	archivepath = FindFirstArchiveinFilepath( filepath );
		if ( archivepath.IsEmpty() ) return OCA_false;
		ziptype zt = Archive::Categorise( archivepath ); 
		if ( zt == NotArchive )	return OCA_false;		// Shouldn't happen!
		if ( ! NewArc(archivepath, zt ) ) return OCA_false;	// Enter the 1st archive
	}

if ( MightFilepathBeWithinChildArchive( filepath ) )	// ie we're going forwards, not backwards
	while (true)
		{	if ( IsWithinThisArchive( filepath ) )  return OCA_true;	// Are we nearly there yet?
			archivepath = GetArc()->Getffs()->GetRootDir()->GetFilepath(); archivepath = archivepath.BeforeLast(wxFILE_SEP_PATH);
			wxString extra; if ( ! filepath.StartsWith( archivepath, &extra ) ) return OCA_false;	// Shouldn't happen!
																				// Extra now contains the unused bit of filepath. 
			wxString nextbit = FindFirstArchiveinFilepath( extra );	// Get the next archive in the path
			if ( nextbit.IsEmpty() )  return OCA_false;
			archivepath += nextbit;
			ziptype zt = Archive::Categorise( archivepath ); 
			if ( zt == NotArchive )	return OCA_false;	// Shouldn't happen!
			if ( ! NewArc(archivepath, zt ) ) return OCA_false; 	// Enter the next archive in the chain
		}

bool flag=true;					// It wasn't further into this archive, so try reversing
while ( true )
	{	flag = Pop();
		if ( IsWithinThisArchive( filepath ) ) return OCA_true;
		if ( ! flag ) return OCA_retry;							// We're out of the old archive. Retry says do it again, to enter any new one
	}
 
return OCA_false;
#else
return OCA_false;
#endif
}

enum NavigateArcResult ArchiveStreamMan::NavigateArchives( wxString filepath )	// If filepath is in an archive, goto the archive. Otherwise don't
{
#if wxVERSION_NUMBER >= 2504
if ( filepath.IsEmpty() ) return Snafu;

FileData stat( filepath );

if ( ! IsArchive() )							// If we're not already inside an archive, we may need to sort out fulltree mode:
	{	if ( (stat.IsValid() && Archive::Categorise( filepath ) != NotArchive)				// if we're entering an archive,
			|| ( ! stat.IsValid() && ! FindFirstArchiveinFilepath( filepath ).IsEmpty()) )	// or if we're jumping deep into an archive,
				{	MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane(); if ( active->fileview == ISRIGHT ) active = active->partner;
					WasOriginallyFulltree = active->fulltree; active->fulltree = false;		// Turn off fulltree mode if it's on, after storing the current state
				}
	}

if ( stat.IsValid() && filepath != OriginalArchive )	// If this is an ordinary filepath, escape from any archive(s) and return for normal processing
	{ while ( Pop() ) ; return OutsideArc; } 				// The only other reason for stat being valid is if it's the original archive; in which case no point exiting here

enum OCA ans;
do ans = OpenCorrectArchive(filepath);					// Navigate either in, out or shake it all about
	while ( ans == OCA_retry );								// We'll need to retry if we *were* in an archive, and now have to come out to enter a different one
return ( ans==OCA_true ? FoundInArchive : Snafu);
#else
return OutsideArc;
#endif
}

wxString ArchiveStreamMan::ExtractToTempfile()	// Called by FiletypeManager::Openfunction to provide a extracted tempfile
{
#if wxVERSION_NUMBER >= 2504
MyGenericDirCtrl* active = MyFrame::mainframe->GetActivePane(); if ( active==NULL ) return wxEmptyString; 
wxString filepath = active->GetFilePath(); if ( filepath.IsEmpty() ) return filepath;			 // Find the filepath to open

wxArrayString filepaths; filepaths.Add( filepath );			// Although there's only 1 filepath, the real ExtractToTempfile() needs an array
wxArrayString resultingfilepaths = ExtractToTempfile( filepaths );

if ( resultingfilepaths.IsEmpty() ) return wxEmptyString;
return resultingfilepaths[0];											// The resulting filepaths are returned in an array. We know there'll only be one of them
#else
return wxEmptyString;
#endif
}

wxArrayString ArchiveStreamMan::ExtractToTempfile( wxArrayString& filepaths )	// Called by EditorBitmapButton::OnEndDrag etc, to provide extracted tempfile(s)
{
wxArrayString tempfilepaths;
#if wxVERSION_NUMBER >= 2504
if ( ! IsArchive() || arc==NULL || filepaths.IsEmpty() ) return tempfilepaths;

wxFileName tempdirbase;
wxString temppath;
while ( true )																	// Do in a loop in case there's already one there ?!?
	{	if ( ! DirectoryForDeletions::GetUptothemomentDirname( tempdirbase, tempfilecan ) ) return tempfilepaths;	// Make a temporary dir to hold the files
		temppath = tempdirbase.GetFullPath();
		FileData stat( temppath ); if ( stat.IsValid() ) break;
	}

if ( ! arc->Extract( filepaths, temppath, tempfilepaths ) ) tempfilepaths.Clear();	// Extract
#endif
return tempfilepaths;														// Return the extracted tempfilepaths
}


