/////////////////////////////////////////////////////////////////////////////
// Name:        Configure.h
// Purpose:    Configuration
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////
#ifndef CONFIGUREH
#define CONFIGUREH

#include "wx/notebook.h"
#include <wx/wizard.h>

#include "Devices.h"
#include "Tools.h"

class NoConfigWizard;

class MyWizardPage : public wxWizardPage
{
public:
MyWizardPage(wxWizard *parent)	:	wxWizardPage(parent){}
virtual void OnPageDisplayed()=0;		// Abstract, as we certainly don't want any objects,
};

class FirstPage : public MyWizardPage
{
public:
FirstPage(wxWizard *parent, wxWizardPage* Page2, wxWizardPage* Page3); 

    // implement wxWizardPage functions
virtual wxWizardPage* GetPrev() const { return (wxWizardPage*)NULL; }
virtual wxWizardPage* GetNext() const  { return RCFound ? m_Page3 : m_Page2; }	// Return the correct next page, depending on whether resources were found
void OnPageDisplayed(){}

protected:
wxWizardPage *m_Page2, *m_Page3;
bool RCFound;
};

class SecondPage : public MyWizardPage
{
public:
SecondPage(wxWizard *parent,  wxWizardPage* Page3); 

virtual wxWizardPage* GetPrev() const { return (wxWizardPage*)NULL; }
virtual wxWizardPage* GetNext() const  { return m_Page3; }
void OnPageDisplayed();

protected:
wxWizard *parent;
wxWizardPage* m_Page3;
};

class ThirdPage : public MyWizardPage
{
public:
ThirdPage(NoConfigWizard *parent, const wxString& configFP); 

virtual wxWizardPage* GetPrev() const { return (wxWizardPage*)NULL; }
virtual wxWizardPage* GetNext() const  { return (wxWizardPage*)NULL; }
void OnPageDisplayed();
protected:
const wxString configFPath;
};

class NoConfigWizard	:	public wxWizard
{
public:
NoConfigWizard(wxWindow* parent, int id, const wxString& title, const wxString& configFPath );
wxCheckBox* shortcut;				// So the user can ask for a desktop shortcut (not) to be provided if there isn't one already
FirstPage* Page1;
protected:
void OnPageChanged(wxWizardEvent& event) { ((MyWizardPage*)event.GetPage())->OnPageDisplayed(); }
void OnButton(wxCommandEvent& event);
private:
DECLARE_EVENT_TABLE()
};

class MyConfigureDialog		:		public MyButtonDialog
{
protected:
void EndModal(int retCode);
};


class Configure; class ArrayOfToolFolders;

class ChooseStripesDlg	:	public wxDialog
{
public:
void Init( wxColour c0, wxColour c1 ){ dirty = false; Col0 = c0 ; Col1 = c1; DisplayCols();}
wxColour Col0, Col1;
protected:
bool dirty;
void DisplayCols();
void OnChangeButton( wxCommandEvent& event );
DECLARE_EVENT_TABLE()
};


struct DirCtrlTBStruct															// Holds info on the desired extra buttons to load to each Dirview mini-toolbar
	{	wxString bitmaplocation;												// Load the bitmap from here
		int iconNo;																	// Which icon to use
		wxBitmapType icontype;												// Can be wxBITMAP_TYPE_PNG, wxBITMAP_TYPE_XPM or wxBITMAP_TYPE_BMP (other handlers aren't loaded)
		wxString tooltip;
		wxString destination;													// The "url" to go to when clicked  eg The "Home" button will contain /home
	};

WX_DEFINE_ARRAY(DirCtrlTBStruct*, ArrayofDirCtrlTBStructs);

class ToolbarIconsDialog : public TBOrEditorBaseDlg
{
public:
void Init( ArrayofDirCtrlTBStructs* DCTBarray );
virtual int SetBitmap( wxBitmapButton* btn, wxWindow* pa, size_t iconno );	// Returns buttons event id if it works, else false
virtual void AddIcon( wxString& newfilepath, size_t type );
static void LoadDirCtrlTBinfo( ArrayofDirCtrlTBStructs* DCTBarray );
void Save();

protected:
void OnAddButton(wxCommandEvent& WXUNUSED(event));
void OnEditButton(wxCommandEvent& WXUNUSED(event));
void OnDeleteButton(wxCommandEvent& WXUNUSED(event));
void DoUpdateUI(wxUpdateUIEvent& event);
void FillListbox();
void SaveIcons();
void OnInitDialog(wxInitDialogEvent& WXUNUSED(event));

size_t NoOfTools;
wxArrayString Icons;
wxArrayInt Icontypes;

ArrayofDirCtrlTBStructs* DirCtrlTBinfoarray;
wxListBox* list;
DeviceManager* parent;
private:
DECLARE_EVENT_TABLE()
};

class AddToobarIconDlg : public wxDialog
{
public:
void Init( ToolbarIconsDialog* dad, wxString defaultfilepath, wxString defaulttooltip, long icon );

TBOrEditorBaseDlg* grandparent;
wxTextCtrl* Filepath;
wxTextCtrl* Tooltip;
long iconno;

protected:
void OnIconClicked(wxCommandEvent& WXUNUSED(event));
void OnBrowseBtn(wxCommandEvent& WXUNUSED(event));
int CreateButton();

wxBitmapButton* btn;
};

class TBIconSelectDlg : public IconSelectBaseDlg
{
protected:
virtual void DisplayButtonChoices();
};

class ConfigureDisplay			// Encapsulates the Display notebookpage stuff
{
public:
ConfigureDisplay(){ page = NULL; dirty = false; fontchanged = false; InitialiseData(); }
void Init( wxNotebookPage* pg ){ page = pg; InsertData(); }
void InitialiseData();							// Fill the temp vars with the old data. Used initially and also to revert on Cancel
void InsertData();
void OnChangeFont();
void OnDefaultFont();
void OnCancel();
void OnOK();
void OnSelectColours();
void OnConfigureTBIcons();
static void LoadDisplayConfig();
void OnUpdateUI();

bool dirty;

protected:
void Save();

bool tempSHOW_TREE_LINES;
int tempTREE_INDENT;
size_t tempTREE_SPACING;
bool tempSHOWHIDDEN;
bool tempTREAT_SYMLINKTODIR_AS_DIR;
bool tempUSE_STRIPES;
wxColour tempSTRIPE_0;
wxColour tempSTRIPE_1;
bool tempSHOW_TB_TEXTCTRL;
bool tempASK_ON_TRASH;
bool tempASK_ON_DELETE;
wxString tempTRASHCAN;
wxFont tempTREE_FONT;
wxFont tempCHOSEN_TREE_FONT;
bool tempUSE_DEFAULT_TREE_FONT;

wxTextCtrl *fonttext, *defaultfonttext;
bool fontchanged;
wxNotebookPage* page;
};

class ConfigureTerminals		// Encapsulates the Terminals notebookpage stuff
{
public:
ConfigureTerminals(){ page = NULL; dirty = false; fontchanged = false; InitialiseData(); }
void Init( wxNotebookPage* pg ){ page = pg; InsertData(); }
void InitialiseData();							// Fill the temp vars with the old data. Used initially and also to revert on Cancel
void InsertData();
void OnChangeFont();
void OnDefaultFont();
void OnCancel();
void OnOK();
void OnAccept( int id );
static void LoadTerminalsConfig();
static void Save( wxConfigBase* config = NULL );
void OnUpdateUI();

bool dirty;
protected:

wxString tempTERMINAL;
wxString tempTERMINAL_COMMAND;
wxString tempTOOLS_TERMINAL_COMMAND;
wxString tempTERMEM_PROMPTFORMAT;
wxFont tempTERMINAL_FONT;
wxFont tempCHOSEN_TERMINAL_FONT;
bool tempUSE_DEFAULT_TERMINALEM_FONT;

wxComboBox *combo1, *combo2, *combo3;
wxTextCtrl *fonttext, *defaultfonttext;
bool fontchanged;
wxArrayString tempPossibleTerminals, tempPossibleTerminalCommands, tempPossibleToolsTerminalCommands;
wxNotebookPage* page;
};

class ConfigureNet			// Encapsulates the Network notebookpage stuff
{
public:
ConfigureNet(){ page = NULL; dirty = false; InitialiseData(); }
void Init( wxNotebookPage* pg ){ page = pg; InsertData(); }
void InitialiseData();							// Fill the temp vars with the old data. Used initially and also to revert on Cancel
void InsertData();
void OnOK();
static void LoadNetworkConfig();
void AddServer();
void DeleteServer();
void OnUpdateUI();

bool dirty;
bool serverschanged;

protected:
void Save();

wxString tempSMBPATH;
wxString tempSHOWMOUNTFPATH;
wxArrayString tempNFS_SERVERS;

wxListBox* ServerList;
wxTextCtrl* NewServer;
wxTextCtrl* Showmount;
wxTextCtrl* Samba;

wxNotebookPage* page;
};

class ConfigureNotebook;

class ConfigureMisc			// Encapsulates the Misc notebookpage stuff
{
public:
ConfigureMisc(){ page = NULL; dirty = false; InitialiseData(); }
void Init( wxNotebookPage* pg ){ page = pg; InsertData(); }
void InitialiseData();							// Fill the temp vars with the old data. Used initially and also to revert on Cancel
void InsertData();
void OnOK();
static void LoadMiscConfig();
static void LoadStatusbarWidths( int* widths );
void OnConfigureDnD();
void OnConfigureStatusbar();
void OnExportBtn( ConfigureNotebook* note );
void OnUpdateUI();

bool dirty;

protected:
void Save();

size_t tempMAX_NUMBER_OF_UNDOS;
size_t tempMAX_DROPDOWN_DISPLAY;
size_t tempAUTOCLOSE_TIMEOUT;
size_t tempAUTOCLOSE_FAILTIMEOUT;
size_t tempDEFAULT_BRIEFLOGSTATUS_TIME;
size_t tempLINES_PER_MOUSEWHEEL_SCROLL;
sutype tempWHICH_SU;
int tempDnDXTRIGGERDISTANCE;
int tempDnDYTRIGGERDISTANCE;
int tempMETAKEYS_COPY;
int tempMETAKEYS_MOVE;
int tempMETAKEYS_HARDLINK;
int tempMETAKEYS_SOFTLINK;
int  widths[4], tempwidths[4];

wxRadioButton *Ksu, *Gksu, *Gnsu;
wxNotebookPage* page;
};

class ConfigureNotebook	:	public wxNotebook	// Holds the pages of configure dlgs
{
enum pagename { UDCommand, UDShortcuts, ConfigDevicesEditors, ConfigDisplay, ConfigTerminals, ConfigNet, ConfigMisc, NotSelected };
enum UDCPagetype { UDCNew, UDCEdit, UDCDelete };	// To enumerate which of the UDC subpages is loaded

public:
ConfigureNotebook(){}
~ConfigureNotebook(){ delete ConfDisplay; delete ConfTerminals; delete ConfNet; delete ConfMisc; }
void Init( Configure* dad );

bool Dirty();																		// Check there's no unsaved data
class LaunchMiscTools* LMT;

protected:
void InitPage( enum pagename page );
void FillUDCMenusCombo( size_t &item, size_t indent );	// Recursively fill a combo, indenting according to sub-ness of the menu
void OnButtonPressed(wxCommandEvent& event);
void OnUDCLabelComboSelectionChanged(wxCommandEvent& event);
void OnPageChanged(wxNotebookEvent& event);
void OnUpdateUI(wxUpdateUIEvent& WXUNUSED(event));

int m_sel;																			// Stores the UDTool selection during editing
enum pagename currentpage;
enum UDCPagetype UDCPage;
Configure* parent;
wxPanel* panel;
ArrayOfToolFolders* FolderArray;
ArrayOfUserDefinedTools* ToolArray;
wxComboBox* UDCCombo;
wxTextCtrl* UDCFilepath;
wxCheckBox* F10check;

class ConfigureDisplay* ConfDisplay;
class ConfigureTerminals* ConfTerminals;
class ConfigureNet* ConfNet;
class ConfigureMisc* ConfMisc;

bool dirty[NotSelected];
bool editing;

private:
DECLARE_DYNAMIC_CLASS(ConfigureNotebook)
DECLARE_EVENT_TABLE()
};

class Configure				// The class that does all the configure things
{
public:
Configure(){ DirCtrlTBinfoarray = new ArrayofDirCtrlTBStructs; }
~Configure(){	for ( int n = (int)DirCtrlTBinfoarray->GetCount();  n > 0; --n  )
							{ DirCtrlTBStruct* item = DirCtrlTBinfoarray->Item(n-1); delete item; DirCtrlTBinfoarray->RemoveAt(n-1); }
						delete DirCtrlTBinfoarray; 
					}
void LoadConfiguration();
void SaveConfiguration();

static bool LocateResources();		// Where is everything?
static bool LocateResourcesSilently();	// Where is everything? This is just a rough first check, so no error messages/try again
static void CopyConf( const wxString& iniFilepath );	// Copy any master conf file into the local ini
static void SaveDefaultsToIni();	// Make sure there is a valid config file, with reasonably-sensible defaults
static void SaveDefaultSus();		// Save a useful value for WHICH_SU
static void SaveDefaultTerminal( enum distro Distro, double Version );// Save a useful value for TERMINAL etc
static void SaveProbableHAL_Sys_etcInfo(enum distro Distro, double Version);	// Save the probable discovery/loading setup
static void SaveDefaultEditors(enum distro Distro, double Version);	// KWrite &/or GEdit
static void SaveDefaultUDTools();// Starter pack of UDtools
static void SaveDefaultFileAssociations();// Similarly
static void SaveDefaultDirCtrlTBButtons();
static void SaveDefaultMisc();
static void DetectFixedDevices();

static enum distro DeduceDistro( double& VERSION );		// Try to work out which distro & version we're on
static bool FindIni();						// If there's not a valid ini file in $HOME, use a wizard to find/create one
static bool Wizard( const wxString& configFPath );
static void FindPrograms();
static bool TestExistence( wxString name, bool executable = false );	// Is 'name' in $PATH, and optionally is it executable by us

void Configure4Pane();
void ConfigureTooltips();														// Globally turn tooltips on/off, set their delay
void ApplyTooltipPreferences();

ArrayofDirCtrlTBStructs* DirCtrlTBinfoarray;
protected: 
static bool FindRC( wxString& extra );									// Find a valid source of XRC files
static bool FindBitmaps( wxString& extra );							// Find a valid source of bitmaps
static bool FindHelp( wxString& extra );								// Find a valid source of Help

static bool CheckForConfFile( wxString& conf );					// Is there a 4Pane.conf file anywhere, from which to load some config data?

void LoadMisc();																	// Load the one-off things like Filter history
void SaveMisc();																	// Save the one-off things like Filter history

enum helpcontext CallersHelpcontext;									// Stores the original value of the help context, to be replaced on finishing config
};



#endif
    // CONFIGUREH


