/////////////////////////////////////////////////////////////////////////////
// Name:        Mounts.h
// Purpose:    Mounting partitions
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////
#ifndef MOUNTSH
#define MOUNTSH

#include "wx/wx.h"

class MountPartitionDialog	:	public wxDialog	// Select partition & mount-point
{
public:
MountPartitionDialog(){}
void Init( DeviceAndMountManager* dad,  wxString device = wxEmptyString );

wxComboBox* PartitionsCombo;
wxTextCtrl* FstabMountptTxt;
wxTextCtrl* MountptTxt;

protected:
void OnBrowseButton(wxCommandEvent& WXUNUSED(event));
void OnButtonPressed(wxCommandEvent& event){ EndModal( event.GetId() ); }
void OnSelectionChanged(wxCommandEvent& WXUNUSED(event)){ DisplayMountptForPartition(); }
void DisplayMountptForPartition( bool GetDataFromMtab=false );		// This enters the correct mountpt for the selected partition, in the fstab version of the dialog

DeviceAndMountManager* parent;

private:
DECLARE_EVENT_TABLE()
};

class MountSambaDialog	:	public MountPartitionDialog
{
public:
MountSambaDialog(){}
~MountSambaDialog(){ for ( int n = (int)NetworkArray.GetCount();  n > 0; --n  ) { NetworkStruct* item = NetworkArray.Item(n-1); delete item; NetworkArray.RemoveAt(n-1); } }
bool Init();
void OnUpdateUI(wxUpdateUIEvent& event);

wxComboBox* IPCombo;
wxComboBox* HostnameCombo;
wxComboBox* SharesCombo;
wxTextCtrl* Username;
wxTextCtrl* Password;
wxStaticText* AlreadyMounted;

bool InFstab, IsMounted; wxString AtMountPt, FstabMt;		// These hold data for any share that is in fstab, and/or is already mounted 

protected:
bool FindData();
bool GetSharesForServer( wxArrayString& array, wxString& hostname, bool findprinters = false );	// Return  hostname's available shares
void LoadSharesForThisServer( int sel );								// Fill the relevant combo with available shares for the displayed server
void GetMountptForShare();													// Enter into MountptTxt any known mountpoint for the selection in SharesCombo
void OnSelectionChanged(wxCommandEvent& event);

ArrayofNetworkStructs NetworkArray;
private:
DECLARE_EVENT_TABLE()
};

class MountLoopDialog	:	public MountSambaDialog		// For mounting iso-images
{
public:
MountLoopDialog(){}
void Init();
wxTextCtrl* ImageTxt;

protected:
void OnIsoBrowseButton(wxCommandEvent& WXUNUSED(event));

void OnSelectionChanged(wxCommandEvent& WXUNUSED(event)){ DisplayMountptForImage(); }
void DisplayMountptForImage();											// This enters the correct mountpt for the selected Image if it's in fstab

private:
DECLARE_EVENT_TABLE()
};

class MountNFSDialog	:	public MountSambaDialog
{
public:
MountNFSDialog(){}
bool Init();

protected:
bool FindData();																	// Query the network to find NFS servers
void OnAddServerButton(wxCommandEvent& WXUNUSED(event));
void LoadExportsForThisServer( int sel );							// Fill the relevant combo with available exports for the displayed server
bool GetExportsForServer( wxArrayString& exports, wxString& server );	// Return server's available exports
void OnSelectionChanged(wxCommandEvent& WXUNUSED(event));
void OnGetMountptForShare(wxCommandEvent& WXUNUSED(event)){ GetMountptForShare(); }
void GetMountptForShare();												// Enter into MountptTxt any known mountpoint for the selection in SharesCombo

private:
DECLARE_EVENT_TABLE()
};

class UnMountPartitionDialog	:	public MountPartitionDialog		// Unmount a mounted partition
{
public:
UnMountPartitionDialog(){}
void Init( DeviceAndMountManager* dad );	// Everything else uses the base class methods & variables.  Init is different as it enters the Mounted partitions, not the unmounted
};

class UnMountSambaDialog	:	public MountPartitionDialog		// Unmount a mounted samba share
{
public:
UnMountSambaDialog(){}
~UnMountSambaDialog(){ for ( int n = (int)Mntarray.GetCount();  n > 0; --n  )	{ PartitionStruct* item = Mntarray.Item(n-1); delete item; Mntarray.RemoveAt(n-1); }}
size_t Init();
void DisplayMountptForShare();											// Show mountpt corresponding to a share or export
void OnSelectionChanged(wxCommandEvent& WXUNUSED(event)){ DisplayMountptForShare(); }
void SearchForNetworkMounts();											// Scans mtab for established NSF & samba mounts

ArrayofPartitionStructs Mntarray;
bool IsSamba;																		// This flags whether the currently-displayed mount is samba or nfs
private:
DECLARE_EVENT_TABLE()
};

#endif
    // MOUNTSH

